

# Standard
import itertools as itt
import warnings

# Common
import numpy as np
import scipy.sparse as scpsp
from matplotlib.tri import Triangulation as mplTri


_DRESERVED_KEYS = {
    'dgroup': ['lref', 'ldata'],
    'dref': ['ldata', 'group', 'size', 'ind'],
    'dref_static': [],
    'ddata': ['ref', 'group', 'shape', 'data'],
    'dobj': [],
}


_DDEF_PARAMS = {
    'ddata': {
        'source': (str, 'unknown'),
        'dim':    (str, 'unknown'),
        'quant':  (str, 'unknown'),
        'name':   (str, 'unknown'),
        'units':  (str, 'a.u.'),
    },
    'dobj': {
    },
}


_DATA_NONE = False


# #############################################################################
# #############################################################################
#                           Generic
# #############################################################################


def _check_which(ddata=None, dobj=None, which=None, return_dict=None):
    """ Check which in ['data'] + list(self._dobj.keys() """

    # --------------
    # Check inputs

    if return_dict is None:
        return_dict = True

    # Trivial case
    if len(ddata) == 0 and len(dobj) == 0:
        if return_dict is True:
            return None, None
        else:
            return

    # which ('data', or keys of dobj)
    if which is None:
        if len(dobj) == 0:
            which = 'data'
        elif len(dobj) == 1:
            which = list(dobj.keys())[0]

    c0 = which in ['data'] + list(dobj.keys())
    if not c0:
        msg = (
            "Please specify whether to sort:\n"
            + "\t- 'data': the content of self.ddata\n\t- "
            + "\n\t- ".join([
                "'{0}': the content of self.dobj['{0}']".format(k0)
                for k0 in dobj.keys()
            ])
            + "\nProvided:\n\t- {}".format(which)
        )
        raise Exception(msg)

    if return_dict is True:
        if which == 'data':
            dd = ddata
        else:
            dd = dobj[which]
        return which, dd
    else:
        return which


def _check_conflicts(dd=None, dd0=None, dd_name=None):
    """ Detect conflict with existing entries
    """
    dupdate = {}
    dconflict = {}
    for k0, v0 in dd.items():
        if k0 not in dd0.keys():
            continue
        # conflicts
        lk = set(v0.keys()).intersection(dd0[k0].keys())
        lk = [
            kk for kk in lk
            if not (
                isinstance(v0[kk], dd0[k0][kk].__class__)
                and (
                    (
                        isinstance(v0[kk], np.ndarray)
                        and np.allclose(v0[kk], dd0[k0][kk], equal_nan=True)
                    )
                    or (
                        scpsp.issparse(v0[kk])
                        and np.allclose(
                            v0[kk].data, dd0[k0][kk].data, equal_nan=True,
                        )
                    )
                    or (
                        v0[kk] == dd0[k0][kk]
                    )
                )
            )
        ]
        if len(lk) > 0:
            dconflict[k0] = lk
        # updates
        lk = [
            kk for kk in dd0[k0].keys()
            if kk not in v0.keys() and kk not in ['ldata', 'size']
        ]
        if len(lk) > 0:
            dupdate[k0] = lk

    # Conflicts => Exception
    if len(dconflict) > 0:
        msg = (
            "Conflicts with pre-existing values found in {}:\n".format(dd_name)
            + "\n".join([
                f"\t- {dd_name}['{k0}']: {v0}"
                for k0, v0 in dconflict.items()
            ])
        )
        raise Exception(msg)

    # Updates => Warning
    if len(dupdate) > 0:
        msg = (
            "\nExisting {} keys will be overwritten:\n".format(dd_name)
            + "\n".join([
                f"\t- {dd_name}['{k0}']: {v0}"
                for k0, v0 in dupdate.items()
            ])
        )
        warnings.warn(msg)


def _check_remove(key=None, dkey=None, name=None):
    c0 = isinstance(key, str) and key in dkey.keys()
    c1 = (
        isinstance(key, list)
        and all([isinstance(kk, str) and kk in dkey.keys() for kk in key])
    )
    if not (c0 or c1):
        msg = (
            """
            Removed param must be a str already in self.d{}
            It can also be a list of such
            \t- provided: {}
            \t- already available: {}
            """.format(name, key, sorted(dkey.keys()))
        )
        raise Exception(msg)
    if c0:
        key = [key]
    return key


# #############################################################################
# #############################################################################
#                           Removing routines
# #############################################################################

def _remove_group(
    group=None, dgroup0=None, dref0=None, ddata0=None,
    dref_static0=None,
    dobj0=None,
    allowed_groups=None,
    reserved_keys=None,
    ddefparams_data=None,
    ddefparams_obj=None,
    data_none=None,
    max_ndim=None,
):
    """ Remove a group (or list of groups) and all associated ref, data """
    if group is None:
        return dgroup0, dref0, ddata0
    group = _check_remove(key=group, dkey=dgroup0, name='group')

    # Remove groups and orphan ref and data
    for k0 in groups:
        for k1 in dgroup0['lref']:
            del dref0[k1]
        lkdata = [k1 for k1, v1 in ddata0.items() if v1['group'] == (k0,)]
        for kk in lkdata:
            del ddata0[kk]
        del dgroup0[k0]

    # Double-check consistency
    return _consistency(
        ddata=None, ddata0=ddata0,
        dref=None, dref0=dref0,
        dref_static=None, dref_static0=dref_static0,
        dobj=None, dobj0=dobj0,
        dgroup=None, dgroup0=dgroup0,
        allowed_groups=allowed_groups,
        reserved_keys=reserved_keys,
        ddefparams_data=ddefparams_data,
        ddefparams_obj=ddefparams_obj,
        data_none=data_none,
        max_ndim=max_ndim,
    )


def _remove_ref(
    key=None,
    dgroup0=None, dref0=None, ddata0=None,
    dref_static0=None,
    dobj0=None,
    propagate=None,
    allowed_groups=None,
    reserved_keys=None,
    ddefparams_data=None,
    ddefparams_obj=None,
    data_none=None,
    max_ndim=None,
):
    """ Remove a ref (or list of refs) and all associated data """
    if key is None:
        return group0, dref0, ddata0
    key = _check_remove(
        key=key, dkey=dref0, name='ref',
    )

    for k0 in key:
        # Remove orphan ddata
        for k1 in dref0[k0]['ldata']:
            del ddata0[k1]

        # Remove ref from dgroup['lref']
        for k1 in dgroup0.keys():
            if k0 in dgroup0[k1]['lref']:
                dgroup0[k1]['lref'].remove(k0)
        del dref0[k0]

    # Propagate upward
    if propagate is True:
        lg = [k0 for k0 in dgroup0.keys() if len(dgroup0['lref']) == 0]
        for gg in lg:
            del dgroup0[gg]

    # Double-check consistency
    return _consistency(
        ddata=None, ddata0=ddata0,
        dref=None, dref0=dref0,
        dref_static=None, dref_static0=dref_static0,
        dobj=None, dobj0=dobj0,
        dgroup=None, dgroup0=dgroup0,
        allowed_groups=allowed_groups,
        reserved_keys=reserved_keys,
        ddefparams_data=ddefparams_data,
        ddefparams_obj=ddefparams_obj,
        data_none=data_none,
        max_ndim=max_ndim,
    )


def _remove_ref_static(
    key=None,
    which=None,
    propagate=None,
    dref_static0=None,
    ddata0=None,
    dobj0=None,
):
    """ Remove a static ref (or list) or a whole category

    key os provided:
        => remove only the desired key(s)
            works only if key is not used in ddata and dobj

    which is provided:
        => treated as param, the whole category of ref_static is removed
            if propagate, the parameter is removed from ddata and dobj
    """

    lc = [
        key is not None,
        which is not None,
    ]
    if np.sum(lc) != 1:
        msg = "Please provide either key xor which!"
        raise Exception(msg)

    if key is not None:
        if isinstance(key, str):
            key = [key]

        lk0 = [
            k0 for k0, v0 in dref_static0.items()
            if all([kk in v0.keys() for kk in key])
        ]
        if len(lk0) != 1:
            msg = (
                "No / several matches for '{}' in ref_static:\n".format(key)
                + "\n".join([
                    "\t- dref_static[{}][{}]".format(k0, key) for k0 in lk0
                ])
            )
            raise Exception(msg)
        k0 = lk0[0]
        key = _check_remove(
            key=key,
            dkey=dref_static0[k0],
            name='ref_static[{}]'.format(k0),
        )

        # Make sure key is not used (condition for removing)
        for kk in key:
            lk1 = [
                k1 for k1, v1 in ddata0.items()
                if kk == v1.get(k0)
            ]
            lk2 = [
                k1 for k1, v1 in dobj0.items()
                if any([kk == v2.get(k0) for v2 in v1.values()])
            ]
            if len(lk1) > 0 or len(lk2) > 0:
                msg = (
                    "Provided ref_static key ({}) is used in:\n".format(kk)
                    + "\n".join(
                        ["\t- self.ddata['{}']".format(k1) for k1 in lk1]
                        + [
                            "\t- self.dobj['{}']['{}']".format(k2, k0)
                            for k2 in lk2
                        ]
                    )
                )
                raise Exception(msg)
            del dref_static0[k0][kk]

    elif which is not None:
        if which not in dref_static0.keys():
            msg = (
                "Provided which not in dref_static.keys():\n"
                + "\t- Available: {}\n".format(sorted(dref_static0.keys()))
                + "\t- Provided: {}".format(which)
            )
            raise Exception(msg)
        del dref_static0[which]

        # Propagate (delete as partam in ddata and dobj)
        if propagate is None:
            propagate = True

        if propagate is True:
            # ddata
            if which in list(ddata0.values())[0].keys():
                _remove_param(dd=ddata0, dd_name='ddata', param=which)

            # dobj0
            for k0 in dobj0.keys():
                if which in list(dobj0[k0].values())[0].keys():
                    _remove_param(
                        dd=dobj0[k0],
                        dd_name="ddobj['{}']".format(k0),
                        param=which,
                    )


def _remove_data(
    key=None,
    dgroup0=None, dref0=None, ddata0=None,
    dref_static0=None,
    dobj0=None,
    propagate=None,
    allowed_groups=None,
    reserved_keys=None,
    ddefparams_data=None,
    ddefparams_obj=None,
    data_none=None,
    max_ndim=None,
):
    """ Remove a ref (or list of refs) and all associated data """
    if key is None:
        return group0, dref0, ddata0
    key = _check_remove(
        key=key, dkey=ddata0, name='data',
    )

    for k0 in key:
        # Remove key from dgroup['ldata'] and dref['ldata']
        for k1 in dgroup0.keys():
            if k0 in dgroup0[k1]['ldata']:
                dgroup0[k1]['ldata'].remove(k0)
            if k0 in dref0[k1]['ldata']:
                dref0[k1]['ldata'].remove(k0)
        del ddata0[k0]

    # Propagate upward
    if propagate is True:
        lk = [k0 for k0 in dgroup0.keys() if len(dgroup0['ldata']) == 0]
        for kk in lk:
            del dgroup0[kk]
        lk = [k0 for k0 in dref0.keys() if len(dref0['ldata']) == 0]
        for kk in lk:
            del dref0[kk]

    # Double-check consistency
    return _consistency(
        ddata=None, ddata0=ddata0,
        dref=None, dref0=dref0,
        dref_static=None, dref_static0=dref_static0,
        dobj=None, dobj0=dobj0,
        dgroup=None, dgroup0=dgroup0,
        allowed_groups=allowed_groups,
        reserved_keys=reserved_keys,
        ddefparams_data=ddefparams_data,
        ddefparams_obj=ddefparams_obj,
        data_none=data_none,
        max_ndim=max_ndim,
    )


def _remove_obj(
    key=None,
    which=None,
    dobj0=None,
    ddata0=None,
    dref0=None,
    dref_static0=None,
    dgroup0=None,
    allowed_groups=None,
    reserved_keys=None,
    ddefparams_data=None,
    ddefparams_obj=None,
    data_none=None,
    max_ndim=None,
):

    # ------------
    # Check inputs

    lc = [
        key is not None,
        which is not None,
    ]
    if np.sum(lc) != 1:
        msg = "Please provide either key xor which!"
        raise Exception(msg)

    if key is not None:
        # key => delete list of obj
        if isinstance(key, str):
            key = [key]

        lk0 = [
            k0 for k0, v0 in dobj0.items()
            if all([kk in v0.keys() for kk in key])
        ]
        if len(lk0) != 1:
            msg = (
                "No / several matches for '{}' in dobj:\n".format(key)
                + "\n".join([
                    "\t- dobj[{}][{}]".format(k0, key) for k0 in lk0
                ])
            )
            raise Exception(msg)
        k0 = lk0[0]
        key = _check_remove(
            key=key,
            dkey=dobj0[k0],
            name='dobj[{}]'.format(k0),
        )
        for kk in set(key).intersection(dobj0[k0].keys()):
            del dobj0[k0][kk]

    elif which is not None:
        if which not in dobj0.keys():
            msg = (
                "Provided which is not a valid self.dobj.keys()!\n"
                + "\t- provided: {}\n".format(which)
                + "\t- available: {}\n".format(sorted(dobj0.keys()))
            )
            raise Exception(msg)

        del dobj0[which]

    return _consistency(
        ddata=None, ddata0=ddata0,
        dref=None, dref0=dref0,
        dref_static=None, dref_static0=dref_static0,
        dobj=None, dobj0=dobj0,
        dgroup=None, dgroup0=dgroup0,
        allowed_groups=allowed_groups,
        reserved_keys=reserved_keys,
        ddefparams_data=ddefparams_data,
        ddefparams_obj=ddefparams_obj,
        data_none=data_none,
        max_ndim=max_ndim,
    )


# #############################################################################
# #############################################################################
#                           dgroup
# #############################################################################


def _check_dgroup(dgroup=None, dgroup0=None, allowed_groups=None):
    """ dgroup must be
    - str: turned to list
    - list of str
    - dict of dict
    """

    # ----------------
    # Trivial case
    if dgroup in [None, {}]:
        return {}

    # ----------------
    # Check conformity
    c0 = isinstance(dgroup, str) and dgroup not in dgroup0.keys()
    c1 = (
        isinstance(dgroup, list)
        and all([
            isinstance(gg, str) and gg not in dgroup0.keys() for gg in dgroup
        ])
    )
    c2 = (
        isinstance(dgroup, dict)
        and all([
            isinstance(k0, str)
            and k0 not in dgroup0.keys()
            and isinstance(v0, dict)
            and all([
                isinstance(k1, str)
                and k1 in _DRESERVED_KEYS['dgroup']
                and isinstance(v1, list)
                and all([isinstance(v2, str) for v2 in v1])
                for k1, v1 in v0.items()
            ])
            for k0, v0 in dgroup.items()
        ])
    )
    if not (c0 or c1 or c2):
        msg = (
            """
            Added group must be either:
            \t- str: not already in self.dgroup
            \t- list of str: each not already in self.dgroup
            \t- dict: each key not already in self.dgroup, each value a dict
            \t- allowed keys in values are:
            \t\t- {}
            You provided:
            \t- {}
            Already available in self.dgroup:
            {}
            """.format(
                sorted(_DRESERVED_KEYS['dgroup']),
                dgroup,
                '\t- ' + '\n\t- '.join(sorted(dgroup0.keys())),
            )
        )
        raise Exception(msg)

    # Check compliance with allowed groups, if any
    if allowed_groups is not None:
        if c0:
            lg = [dgroup] if dgroup not in allowed_groups else []
        elif c1:
            lg = [k0 for k0 in dgroup if k0 not in allowed_groups]
        else:
            lg = [k0 for k0 in dgroup > keys() if k0 not in allowed_groups]
        if len(lg) > 0:
            msg = (
                """
                The following group names are not allowed:
                {}

                Only the following group names are allowed:
                {}
                """.format(
                    '\t- ' + '\n\t- '.join(lg),
                    '\t- ' + '\n\t- '.join(allowed_groups),
                )
            )
            raise Exception(msg)

    # Convert if necessary
    if c0:
        dgroup = {dgroup: {'lref': [], 'ldata': []}}
    elif c1:
        dgroup = {k0: {'lref': [], 'ldata': []} for k0 in dgroup}
    else:
        dgroup = {k0: {'lref': [], 'ldata': []} for k0 in dgroup.keys()}

    return dgroup


# #############################################################################
# #############################################################################
#                           dref_static
# #############################################################################


def _check_dref_static(
    dref_static=None, dref_static0=None,
):
    """ Check and format dref_staytic

    dref_static can be:
        - dict

    """

    # ----------------
    # Trivial case
    if dref_static in [None, {}]:
        return {}

    # ----------------
    # Check conformity

    c0 = (
        isinstance(dref_static, dict)
        and all([
            isinstance(k0, str)
            and isinstance(v0, dict)
            and all([
                isinstance(k1, str)
                and isinstance(v1, dict)
                for k1, v1 in v0.items()
            ])
            for k0, v0 in dref_static.items()
        ])
    )

    # Raise exception if non-conformity
    if not c0:
        msg = (
            """
            Arg dref_static must be a dict of the form:
            dict(
                'type0': {'k0': {...},
                          'k1': {...}},
                'type1': {'k0': {...},
                          'k1': {...}},
            )
            """
            +
            """
            Provided:
            {}
            """.format(dref_static)
        )
        raise Exception(msg)

    # raise except if conflict with existing entry
    dupdate = {}
    dconflict = {}
    for k0, v0 in dref_static.items():
        lkout = ['nb. data']
        if k0 == 'ion':
            lkout += ['ION', 'charge', 'element']
        if k0 not in dref_static0.keys():
            continue

        for k1, v1 in v0.items():
            if k1 not in dref_static0[k0].keys():
                continue
            # conflicts
            lk = set(v1.keys()).intersection(dref_static0[k0][k1].keys())
            lk = [kk for kk in lk if v1[kk] != dref_static0[k0][k1][kk]]
            if len(lk) > 0:
                dconflict[k0] = (k1, lk)
            # updates
            lk = [
                kk for kk in dref_static0[k0][k1].keys()
                if kk not in v1.keys()
                and kk not in lkout
                and 'nb. ' not in kk
            ]
            if len(lk) > 0:
                dupdate[k0] = (k1, lk)

    # Conflicts => Exception
    if len(dconflict) > 0:
        msg = (
            "The following dref_static keys are conflicting existing values:\n"
            + "\n".join([
                "\t- dref_static['{}']['{}']: {}".format(k0, v0[0], v0[1])
                for k0, v0 in dconflict.items()
            ])
        )
        raise Exception(msg)

    # Updates => Warning
    if len(dupdate) > 0:
        msg = (
            "\nThe following existing dref_static keys will be forgotten:\n"
            + "\n".join([
                "\t- dref_static['{}']['{}']: {}".format(k0, v0[0], v0[1])
                for k0, v0 in dupdate.items()
            ])
        )
        warnings.warn(msg)

    # ------------------
    # Check element / ion / charge
    _check_elementioncharge_dict(dref_static=dref_static)

    return dref_static


# #############################################################################
# #############################################################################
#                           dref
# #############################################################################


class DataRefException(Exception):

    def __init__(self, ref=None, data=None):
        msg = (
            """
            To be a valid reference for {}, provided data must be either:
            \t- np.ndarray:  of dimension 1 with increasing values
            \t- list, tuple: convertible to the above
            \t- dict / other class: used for meshes

            You provided:
            \t- {}

            """.format(ref, data)
        )
        self.message = msg


def _check_dataref(data=None, key=None):
    """ Check the conformity of data to be a valid reference """

    # if not array
    # => try converting or get class (dict, mesh...)
    group = None
    if not isinstance(data, np.ndarray):
        if isinstance(data, list) or isinstance(data, tuple):
            try:
                data = np.array(data)
                size = data.size
            except Exception as err:
                raise DataRefException(ref=key, data=data)
        else:
            try:
                data, size = _check_mesh_temp(data=data, key=key)
                if len(size) == 1:
                    size = size[0]
                group = 'mesh2d'
            except Exception as err:
                size = data.__class__.__name__

    # if array => check unique (unique + sorted)
    if isinstance(data, np.ndarray):
        if data.ndim != 1:
            raise DataRefException(ref=key, data=data)

        datau = np.unique(data)
        if not (datau.size == data.size and np.allclose(datau, data)):
            raise DataRefException(ref=key, data=data)
        size = data.size

    return data, size, group


def _check_dref(
    dref=None, dref0=None, dgroup0=None, ddata0=None, allowed_groups=None,
):
    """ Check and format dref

    dref can be:
        - dict

    If some groups are not already on dgroup0
        => completes dgroups0

    If some data is provided
        => returns ddata to be added

    Also think about meshes !!!
    """

    # ----------------
    # Trivial case
    if dref in [None, {}]:
        return {}, None, None

    # ----------------
    # Check conformity
    ngroup = len(dgroup0)
    if ngroup == 1:
        groupref = list(dgroup0.keys())[0]

    # Basis
    # lk_opt = ['ldata', 'size', 'group', 'data']
    c0 = isinstance(dref, dict)
    lc = [
        k0 for k0, v0 in dref.items()
        if not (
            isinstance(k0, str)
            # and k0 not in dref0.keys()
            and (
                (
                    ngroup == 1
                    and (
                        type(v0) in [np.ndarray, list, tuple]
                        or (
                            isinstance(v0, dict)
                            and all([isinstance(ss, str) for ss in v0.keys()])
                            and ('size' in v0.keys() or 'data' in v0.keys())
                        )
                    )
                )
                or (
                    (ngroup == 0 or ngroup > 1)
                    and isinstance(v0, dict)
                    and all([isinstance(ss, str) for ss in v0.keys()])
                    and ('size' in v0.keys() or 'data' in v0.keys())
                    and (
                        'group' in v0.keys()
                        or (
                            'data' in v0.keys()
                            and isinstance(v0['data'], dict)
                        )
                    )
                )
            )
        )
    ]

    # Raise exception if non-conformity
    if not (c0 and len(lc) == 0):
        msg = (
            """
            Arg dref must be a dict of the form:
            {
                'ref0': {'group': str, 'size': int, ...},       (A)
                'ref1': {'group': str, 'data': np.array, ...},  (B)
                'ref2': {'data': np.array, ...},                (C)
                ...
                'refn': np.array,                               (D)
            }

            Where:
                - each 'refi' is a unique str identifier
                - (A) & (B): 'group' is provided as well as 'size' of 'data'
                - (C): 'group' is not provided if len(self.dgroup) == 1
                - (D): only the data array is provided if len(self.dgroup) == 1

            Non-conform refs
            """
            + '\t- ' + '\n\t- '.join(lc)
        )
        raise Exception(msg)

    # -----------------------
    # Make sure all are dict
    for k0, v0 in dref.items():
        if not isinstance(v0, dict):
            dref[k0] = {'data': v0}

    # -----------------------
    # raise except if conflict with existing entry
    _check_conflicts(dd=dref, dd0=dref0, dd_name='dref')

    # ----------------
    # Add size / data if relevant
    ddata_add = {
        k0: {'data': None}
        for k0, v0 in dref.items()
        if 'data' in v0.keys() and k0 not in ddata0.keys()
    }
    for k0, v0 in dref.items():
        if 'data' in v0.keys():
            data, dref[k0]['size'], group = _check_dataref(
                data=v0['data'], key=k0,
            )
            if k0 in ddata_add.keys():
                ddata_add[k0]['data'] = data
                ddata_add[k0]['ref'] = (k0,)
                ddata_add[k0].update({
                    k1: v1 for k1, v1 in v0.items()
                    if k1 not in ['group', 'size', 'ldata']
                })
            if group is not None and dref.get('group') is None:
                dref[k0]['group'] = group

    # Make sure, if ngroup != 1, that NOW all refs have a group
    if ngroup != 1:
        lerr = [k0 for k0, v0 in dref.items() if v0.get('group') is None]
        if len(lerr) > 0:
            msg = "Some groups remain ambiguous!:\n{}".format(lerr)
            raise Exception(msg)

    # ----------------
    # Convert and/or add group if necessary
    for k0, v0 in dref.items():
        if v0.get('group') is None:
            dref[k0]['group'] = groupref

    # Add missing groups
    lgroups = sorted(set([
        v0['group'] for v0 in dref.values()
        if 'group' in v0.keys() and v0['group'] not in dgroup0.keys()
    ]))

    dgroup_add = None
    if len(lgroups) > 0:
        dgroup_add = _check_dgroup(
            lgroups, dgroup0=dgroup0, allowed_groups=allowed_groups,
        )

    # get rid of extra keys
    dref = {
        k0: {k1: v1 for k1, v1 in v0.items() if k1 in _DRESERVED_KEYS['dref']}
        for k0, v0 in dref.items()
    }
    return dref, dgroup_add, ddata_add


# #############################################################################
# #############################################################################
#               ddata - special case: meshes
# #############################################################################


def _get_RZ(arr, name=None, shapeRZ=None):
    if arr.ndim == 1:
        if np.any(np.diff(arr) <= 0.):
            msg = "Non-increasing {}".format(name)
            raise Exception(msg)
    else:
        lc = [np.all(np.diff(arr[0, :])) > 0.,
              np.all(np.diff(arr[:, 0])) > 0.]
        if np.sum(lc) != 1:
            msg = "Impossible to know {} dimension!".format(name)
            raise Exception(msg)
        if lc[0]:
            arr = arr[0, :]
            if shapeRZ[1] is None:
                shapeRZ[1] = name
            if shapeRZ[1] != name:
                msg = "Inconsistent shapeRZ"
                raise Exception(msg)
        else:
            arr = arr[:, 0]
            if shapeRZ[0] is None:
                shapeRZ[0] = name
            if shapeRZ[0] != name:
                msg = "Inconsistent shapeRZ"
                raise Exception(msg)
    return arr, shapeRZ


def _duplicates(arr, arru, nn, name=None, msg=None):
    msg += (
        "  Duplicate {}: {}\n".format(name, nn - arru.shape[0])
        + "\t- {}.shape: {}\n".format(name, arr.shape)
        + "\t- unique shape: {}".format(arru.shape)
    )
    return msg


def _check_trimesh_conformity(nodes, faces, key=None):
    nnodes = nodes.shape[0]
    nfaces = faces.shape[0]

    # Test for duplicates
    nodesu = np.unique(nodes, axis=0)
    facesu = np.unique(faces, axis=0)
    lc = [nodesu.shape[0] != nnodes,
          facesu.shape[0] != nfaces]
    if any(lc):
        msg = "Non-valid mesh ddata[{0}]: \n".format(key)
        if lc[0]:
            msg = _duplicates(nodes, nodesu, nnodes, name='nodes', msg=msg)
        if lc[1]:
            msg = _duplicates(faces, facesu, nfaces, name='faces', msg=msg)
        raise Exception(msg)

    # Test for unused nodes
    facesu = np.unique(facesu)
    c0 = np.all(facesu >= 0) and facesu.size == nnodes
    if not c0:
        ino = str([ii for ii in range(0, nnodes) if ii not in facesu])
        msg = "Unused nodes in ddata[{0}]:\n".format(key)
        msg += "    - unused nodes indices: {}".format(ino)
        warnings.warn(msg)

    # Check counter-clockwise orientation
    x, y = nodes[faces, 0], nodes[faces, 1]
    orient = ((y[:, 1] - y[:, 0])*(x[:, 2] - x[:, 1])
              - (y[:, 2] - y[:, 1])*(x[:, 1] - x[:, 0]))

    clock = orient > 0.
    if np.any(clock):
        msg = ("Some triangles not counter-clockwise\n"
               + "  (necessary for matplotlib.tri.Triangulation)\n"
               + "    => {}/{} triangles reshaped".format(clock.sum(), nfaces))
        warnings.warn(msg)
        faces[clock, 1], faces[clock, 2] = faces[clock, 2], faces[clock, 1]
    return faces


def _check_mesh_temp(data=None, key=None):
    # Check if provided data is mesh (as a dict)

    # ------------
    # Check basics
    lmok = ['rect', 'tri', 'quadtri']
    c0 = (
        isinstance(data, dict)
        and all([ss in data.keys() for ss in ['type']])
        and data['type'] in lmok
        and (
            (
                data['type'] == 'rect'
                and all([ss in data.keys() for ss in ['R', 'Z']])
                and isinstance(data['R'], np.ndarray)
                and isinstance(data['Z'], np.ndarray)
                and data['R'].ndim in [1, 2]
                and data['Z'].ndim in [1, 2]
            )
            or (
                data['type'] in ['tri', 'quadtri', 'quad']
                and all([ss in data.keys() for ss in ['nodes', 'faces']])
                and isinstance(data['nodes'], np.ndarray)
                and isinstance(data['faces'], np.ndarray)
                and data['nodes'].ndim == 2
                and data['faces'].ndim == 2
                and data['faces'].dtype == np.int
                and data['nodes'].shape[1] == 2
                and (
                    (
                        data['type'] in ['tri', 'quadtri']
                        and data['faces'].shape[1] == 3
                    )
                    or (
                        data['type'] == 'quad'
                        and data['faces'].shape[1] == 4
                    )
                )
                and np.max(data['faces']) <= data['nodes'].shape[0]
            )
        )
    )
    if not c0:
        msg = (
            """
            A mesh should be a dict of one of the following form:

                dict(
                 'type': 'rect',
                 'R': np.ndarray (with ndim in [1, 2]),
                 'Z': np.ndarray (with ndim in [1, 2]),
                 'shapeRZ': ('R', 'Z') or ('Z', 'R')
                )

                 dict(
                 'type': 'tri' or 'quadtri',
                 'nodes': np.ndarray of shape (N, 2),
                 'faces': np.ndarray of int of shape (N, 3)
                )

                dict(
                 'type': 'quad',
                 'nodes': np.ndarray of shape (N, 2),
                 'faces': np.ndarray of int of shape (N, 4)
                )

            Provided:
            {}
            """.format(data)
        )
        raise Exception(msg)

    # ------------
    # Check per type
    if data['type'] == 'rect':

        shapeRZ = data.get('shapeRZ', [None, None])
        if shapeRZ is None:
            shapeRZ = [None, None]
        else:
            shapeRZ = list(shapeRZ)

        R, shapeRZ = _get_RZ(data['R'], name='R', shapeRZ=shapeRZ)
        Z, shapeRZ = _get_RZ(data['Z'], name='Z', shapeRZ=shapeRZ)
        shapeRZ = tuple(shapeRZ)

        if shapeRZ not in [('R', 'Z'), ('Z', 'R')]:
            msg = "Inconsistent shapeRZ"
            raise Exception(msg)

        def trifind(
            r, z,
            Rbin=0.5*(R[1:] + R[:-1]),
            Zbin=0.5*(Z[1:] + Z[:-1]),
            nR=R.size, nZ=Z.size,
            shapeRZ=shapeRZ
        ):
            indR = np.searchsorted(Rbin, r)
            indZ = np.searchsorted(Zbin, z)
            indR[(r < R[0]) | (r > R[-1])] = -1
            indZ[(z < Z[0]) | (z > Z[-1])] = -1
            return indR, indZ
            # if shapeRZ == ('R', 'Z'):
            #     indpts = indR*nZ + indZ
            # else:
            #     indpts = indZ*nR + indR
            # indout = ((r < R[0]) | (r > R[-1])
            #           | (z < Z[0]) | (z > Z[-1]))
            # indpts[indout] = -1
            # return indpts

        data['R'] = R
        data['Z'] = Z
        data['shapeRZ'] = shapeRZ
        data['nR'] = R.size
        data['nZ'] = Z.size
        data['shape'] = (R.size, Z.size)
        data['trifind'] = trifind
        data['ftype'] = data.get('ftype', 0)

        if data['ftype'] != 0:
            msg = "Linear interpolation not handled yet !"
            raise Exception(msg)

    else:
        # Check mesh conformity for triangulation
        data['faces'] = _check_trimesh_conformity(
            nodes=data['nodes'], faces=data['faces'], key=key
        )

        data['nnodes'] = data['nodes'].shape[0]
        data['nfaces'] = data['faces'].shape[0]
        data['ftype'] = data.get('ftype', 0)

        # Convert 'quad' to 'quadtri' if relevant
        if data['type'] == 'quad':
            # Convert to tri mesh (solution for unstructured meshes)
            faces = np.empty((data['nfaces']*2, 3), dtype=int)
            faces[::2, :] = data['faces'][:, :3]
            faces[1::2, :-1] = data['faces'][:, 2:]
            faces[1::2, -1] = data['faces'][:, 0]
            data['faces'] = faces
            data['type'] = 'quadtri'
            data['ntri'] = 2

            # Re-check mesh conformity
            data['faces'] = _check_trimesh_conformity(
                nodes=data['nodes'], faces=data['faces'], key=key
            )

        # Check ntri
        if data['type'] == 'tri':
            data['ntri'] = 1
        elif 'ntri' not in data.keys():
            msg = (
                """
                For ddata[{}] of type 'quadtri', 'ntri' must be provided
                """.format(key)
            )
            raise Exception(msg)

        # Only triangular meshes so far
        if 'tri' in data['type']:
            if data.get('mpltri', None) is None:
                data['mpltri'] = mplTri(
                    data['nodes'][:, 0],
                    data['nodes'][:, 1],
                    data['faces']
                )
            if not isinstance(data['mpltri'], mplTri):
                msg = (
                    """
                    ddata[{}]['mpltri'] must be a matplotlib Triangulation
                    Provided:
                    {}
                    """.format(key, data['mpltri'])
                )
            assert data['ftype'] in [0, 1]
            if data['ftype'] == 1:
                data['shape'] = (data['nnodes'],)
            else:
                data['shape'] = (int(data['nfaces'] / data['ntri']),)

    return data, data['shape']


# #############################################################################
# #############################################################################
#               ddata - special case: roman to int (SpectralLines)
# #############################################################################


def roman2int(ss):
    """
    :type s: str
    :rtype: int

    source: https://www.tutorialspoint.com/roman-to-integer-in-python
    """
    roman = {
        'I': 1,
        'V': 5,
        'X': 10,
        'L': 50,
        'C': 100,
        'D': 500,
        'M': 1000,
        'IV': 4,
        'IX': 9,
        'XL': 40,
        'XC': 90,
        'CD': 400,
        'CM': 900,
    }
    i = 0
    num = 0
    while i < len(ss):
        if i+1 < len(ss) and ss[i:i+2] in roman:
            num += roman[ss[i:i+2]]
            i += 2
        else:
            num += roman[ss[i]]
            i += 1
    return num


def int2roman(num):
    roman = {
        1000: "M",
        900: "CM",
        500: "D",
        400: "CD",
        100: "C",
        90: "XC",
        50: "L",
        40: "XL",
        10: "X",
        9: "IX",
        5: "V",
        4: "IV",
        1: "I",
    }

    def roman_num(num):
        for r in roman.keys():
            x, y = divmod(num, r)
            yield roman[r] * x
            num -= (r * x)
            if num <= 0:
                break

    return "".join([a for a in roman_num(num)])


# #############################################################################
# #############################################################################
#                           ddata
# #############################################################################


def _check_data(data=None, key=None, max_ndim=None):
    """ Check the conformity of data to be a valid reference """

    # if not array
    # => try converting or get class (dict, mesh...)
    shape = None
    group = None
    c0_array = (
        isinstance(data, np.ndarray)
        or scpsp.issparse(data)
    )
    if not c0_array:
        if isinstance(data, list) or isinstance(data, tuple):
            c0 = (
                all([hasattr(oo, '__iter__') for oo in data])
                and len(set([len(oo) for oo in data])) != 1
            )
            c1 = (
                all([hasattr(oo, '__iter__') for oo in data])
                and len(set([len(oo) for oo in data])) == 1
            )
            if c0:
                data = np.array(data, dtype=object)
                shape = (data.shape[0],)
            elif c1:
                data = np.array(data)
            else:
                try:
                    data = np.array(data)
                    shape = data.shape
                except Exception as err:
                    raise DataRefException(ref=key, data=data)
        else:
            try:
                data, shape = _check_mesh_temp(data=data, key=key)
                group = 'mesh2d'
            except Exception as err:
                shape = data.__class__.__name__

    # if array => check unique (unique + sorted)
    if c0_array and shape is None:
        shape = data.shape

    # Check max_dim if any
    if c0_array and max_ndim is not None:
        if data.ndim > max_ndim:
            msg = (
                """
                Provided data for ddata[{}] has too many dimensions!
                - ndim:     {}
                - max_ndim: {}
                """.format(key, data.ndim, max_ndim)
            )
            raise Exception(msg)

    # Check if valid ref candidate
    if isinstance(data, np.ndarray):
        monotonous = tuple([
            bool(
                np.all(np.diff(data, axis=aa) > 0.)
                or np.all(np.diff(data, axis=aa) < 0.)
            )
            for aa in range(data.ndim)
        ])
    else:
        monotonous = (False,)
    return data, shape, group, monotonous


def _check_ddata(
    ddata=None,
    ddata0=None,
    dref0=None,
    dgroup0=None,
    reserved_keys=None,
    allowed_groups=None,
    data_none=None,
    max_ndim=None,
):

    # ----------------
    # Trivial case
    if ddata in [None, {}]:
        return {}, None, None
    if data_none is None:
        data_none = _DATA_NONE

    # ----------------
    # Check conformity
    nref = len(dref0)
    if nref == 1:
        refref = list(dref0.keys())[0]

    # Basis
    # lk_opt = ['ldata', 'size', 'group', 'data']
    c0 = isinstance(ddata, dict)
    lc = [
        k0 for k0, v0 in ddata.items()
        if not (
            isinstance(k0, str)
            # and k0 not in ddata0.keys()
            and (
                (
                    nref == 1
                    and (
                        isinstance(v0, (np.ndarray, list, tuple))
                        or scpsp.issparse(v0)
                        or (
                            isinstance(v0, dict)
                            and all([isinstance(ss, str) for ss in v0.keys()])
                            and (
                                (
                                    'data' in v0.keys()
                                    and (
                                        v0.get('ref') is None
                                        or isinstance(v0.get('ref'), str)
                                        or isinstance(v0.get('ref'), tuple)
                                        or v0.get('ref') is True
                                    )
                                )
                                or (
                                    data_none is True
                                    and v0.get('data') is None
                                )
                            )
                        )
                    )
                )
                or (
                    (nref == 0 or nref > 1)
                    and isinstance(v0, dict)
                    and all([isinstance(ss, str) for ss in v0.keys()])
                    and (
                        (
                            'data' in v0.keys()
                            and (
                                (
                                    'ref' in v0.keys()
                                    and (
                                        isinstance(v0.get('ref'), str)
                                        or isinstance(v0.get('ref'), tuple)
                                        or v0.get('ref') is True
                                    )
                                )
                                or (
                                    isinstance(v0['data'], dict)
                                    or isinstance(v0.get('ref'), str)
                                    or isinstance(v0.get('ref'), tuple)
                                    or v0.get('ref') in [None, True]
                                )
                            )
                        )
                        or (
                            data_none is True
                            and v0.get('data') is None
                        )
                    )
                )
            )
        )
    ]

    # Raise exception if non-conformity
    if not (c0 and len(lc) == 0):
        msg = (
            """
            Arg ddata must be a dict of the form:
            dict(
                'data0': {'ref': 'ref0', 'data': list, ...},       (A)
                'data1': {'ref': ('ref0', 'ref1'), 'data': np.array, ...},  (B)
                'data2': {'data': np.array, ...},                (C)
                ...
                'datan': np.array,                               (D)
            )

            Where:
                - each 'datai' is a unique str identifier
                - (A) & (B): 'data' is provided as well as 'ref'
                - (C): 'ref' is not provided if len(self.dref) == 1
                - (D): only the data array is provided if len(self.dgroup) == 1

            If ref = True, the data is itself considered a ref

            The following keys do not match the criteria:
            """
            + '\t- '+'\n\t- '.join(lc)
        )
        raise Exception(msg)

    # -----------------------
    # raise except if conflict with existing entry
    _check_conflicts(dd=ddata, dd0=ddata0, dd_name='ddata')

    # ----------------
    # Convert and/or add ref if necessary
    lref_add = None
    for k0, v0 in ddata.items():
        if not isinstance(v0, dict):
            ddata[k0] = {'ref': (refref,), 'data': v0}
        else:
            if v0.get('data') is None:
                continue
            if v0.get('ref') is None:
                if not isinstance(v0['data'], dict):
                    ddata[k0]['ref'] = (refref,)
            elif isinstance(v0['ref'], str):
                ddata[k0]['ref'] = (v0['ref'],)
            elif v0['ref'] is True:
                if k0 not in dref0.keys():
                    if lref_add is None:
                        lref_add = [k0]
                    else:
                        lref_add.append(k0)
                ddata[k0]['ref'] = (k0,)

    # Check data and ref vs shape - and optionnally add to ref if mesh2d
    for k0, v0 in ddata.items():
        if v0.get('data') is not None:
            (
                ddata[k0]['data'], ddata[k0]['shape'],
                group, ddata[k0]['monot']
            ) = _check_data(
                data=v0['data'], key=k0, max_ndim=max_ndim,
            )

            # Check if group / mesh2d
            if group is not None:
                c0 = ddata[k0].get('ref') in [None, (k0,)]
                if not c0:
                    msg = (
                        """
                        ddata[{}]['ref'] is a {}
                          => it should have ref = ({},)
                        """.format(k0, group, k0)
                    )
                    raise Exception(msg)
                ddata[k0]['ref'] = (k0,)
                c0 = (
                    (lref_add is None or k0 not in lref_add)
                    and k0 not in dref0.keys()
                )
                if c0:
                    if lref_add is None:
                        lref_add = [k0]
                    else:
                        lref_add.append(k0)

    # Add missing refs (only in ddata)
    dgroup_add = None
    dref_add = None
    lref = list(itt.chain.from_iterable([
        [
            rr for rr in v0['ref']
            if rr not in dref0.keys() and rr in ddata.keys()
        ]
        for v0 in ddata.values() if (
            'ref' in v0.keys() and v0.get('data') is not None
        )
    ]))

    if lref_add is not None:
        lref += lref_add

    if len(lref) > 0:
        lref = set(lref)
        dref_add = {rr: {'data': ddata[rr]['data']} for rr in lref}
        dref_add, dgroup_add, ddata_dadd = _check_dref(
            dref=dref_add, dref0=dref0, ddata0=ddata0, dgroup0=dgroup0,
            allowed_groups=allowed_groups,
        )

    # Check shape vs ref
    for k0, v0 in ddata.items():
        if v0.get('data') is None:
            continue
        c0 = (
            isinstance(v0['ref'], tuple)
            and all([
                ss in dref0.keys()
                or (dref_add is not None and ss in dref_add.keys())
                for ss in v0['ref']
            ])
        )
        if not c0:
            msg = (
                f"ddata['{k0}']['ref'] contains unknown ref:\n"
                f"\t- ddata['{k0}']['ref'] = {v0['ref']}\n"
                f"\t- dref0.keys() = {sorted(dref0.keys())}\n"
                + "\t- dref_add.keys() = {}".format(
                    None if dref_add is None else sorted(dref_add.keys())
                )
            )
            raise Exception(msg)
        if c0:
            if isinstance(v0['shape'], tuple):
                shaperef = [
                    dref0[rr]['size'] if rr in dref0.keys()
                    else dref_add[rr]['size']
                    for rr in v0['ref']
                ]
                c1 = (
                    len(shaperef) > 1
                    or any([isinstance(ss, tuple) for ss in shaperef])
                )
                if c1:
                    shaperef = np.r_[tuple(shaperef)].ravel()
                shaperef = tuple(shaperef)
                c0 = c0 and shaperef == v0['shape']
            else:
                c0 = v0['ref'] == (k0,)

        # Raise Exception if needed
        if not c0:
            if isinstance(v0['shape'], tuple):
                msg = (
                    """
                    Inconsistent shape vs ref for ddata[{0}]:
                        - ddata['{0}']['ref'] = {1}  ({2})
                        - ddata['{0}']['shape'] = {3}

                    If dict / object it should be its own ref!
                    """.format(k0, v0['ref'], shaperef, v0['shape'])
                )
            else:
                msg = (
                    "ddata[{0}]['ref'] != ({0},)".format(k0)
                    + "\n\t- ddata[{}]['ref'] = {}\n\n".format(k0, v0['ref'])
                    + "... or there might be an issue with:\n"
                    + "\t- type(ddata[{}]['shape']) = {} ({})".format(
                        k0, type(v0['shape']), v0['shape'],
                    )
                )
            raise Exception(msg)

    return ddata, dref_add, dgroup_add


# #############################################################################
# #############################################################################
#                           dobj
# #############################################################################


def _check_dobj(
    dobj=None, dobj0=None,
):

    # ----------------
    # Trivial case
    if dobj in [None, {}]:
        return {}

    # ----------------
    # Check conformity

    # map possible non-conformities
    if not isinstance(dobj, dict):
        msg = (
            "Arg dobj must be a dict!\n"
            "\t- Provided: {}".format(type(dobj))
        )
        raise Exception(msg)

    # Map possible non-conformities
    dc = {}
    for k0, v0 in dobj.items():
        c1 = isinstance(k0, str) and isinstance(v0, dict)
        if not c1:
            dc[k0] = "type(key) != str or type(value) != dict"
            continue

        if k0 not in dobj0.keys():
            lc2 = [k1 for k1 in v0.keys() if not isinstance(k1, str)]
            if len(lc2) > 0:
                dc[k0] = (
                    "The following keys of dobj[{}] are not str:\n".format(k0)
                    + "\n\t- "
                    + "\n\t- ".join(lc2)
                )
                continue
        else:
            lc2 = [
                k1 for k1 in v0.keys()
                if not isinstance(k1, str)
                or k1 in dobj0[k0].keys()
            ]
            if len(lc2) > 0:
                dc[k0] = (
                    "The following keys of dobj[{}] are not str:\n".format(k0)
                    + "\n\t- "
                    + "\n\t- ".join(lc2)
                    + "(or they are already in dobj0[{}]".format(k0)
                )

    # Raise Exception
    if len(dc) > 0:
        msg = (
            "The following keys of dobj are non-conform:\n"
            + "\n\n".join([
                'dobj[{}]: {}'.format(k0, v0) for k0, v0 in dc.items()
            ])
        )
        raise Exception(msg)

    return dobj


# #############################################################################
# #############################################################################
#                           Params
# #############################################################################


def _check_elementioncharge(
    ION=None, ion=None,
    element=None, charge=None,
    warn=None,
):
    """ Specific to SpectralLines """

    if warn is None:
        warn = True

    # Assess if relevant
    lc = [
        ION is not None,
        ion is not None,
        element is not None and charge is not None,
    ]
    if not any(lc):
        if warn is True:
            msg = (
                """
                To determine ION, ion, element and charge, provide either:
                - ION:  {}
                - ion:  {}
                - element and charge: {}, {}
                """.format(ION, ion, element, charge)
            )
            warnings.warn(msg)
        return None, None, None, None

    # Get element and charge from ION if any
    if lc[0] or lc[1]:
        indc = 1
        if (lc[0] and ION[1].islower()) or (lc[1] and ion[1].islower()):
            indc = 2

        # Infer element
        elementi = ION[:indc] if lc[0] else ion[:indc]
        if element is not None and element != elementi:
            msg = (
                """
                Inconsistent ION ({}) vs element ({})
                """.format(element, elementi)
            )
            raise Exception(msg)

        # Infer charge
        if lc[0]:
            chargei = roman2int(ION[indc:]) - 1
        else:
            chargei = int(ion[indc:].replace('+', ''))
        if charge is not None and charge != chargei:
            msg = (
                """
                Inconsistent ION ({}) vs charge ({})
                """.format(charge, chargei)
            )
            raise Exception(msg)
        element = elementi
        charge = chargei
        if lc[0]:
            ioni = '{}{}+'.format(element, charge)
            if lc[1] and ioni != ion:
                msg = (
                    """
                    Inconsistent ION ({}) vs ion ({})
                    """.format(ION, ion)
                )
                raise Exception(msg)
            ion = ioni

        elif lc[1]:
            IONi = '{}{}'.format(element, int2roman(charge+1))
            if lc[0] and IONi != ION:
                msg = (
                    """
                    Inconsistent ion ({}) vs ION ({})
                    """.format(ion, ION)
                )
                raise Exception(msg)
            ION = IONi

    # ion provided -> element and charge
    elif lc[2]:
        ioni = '{}{}+'.format(element, charge)
        IONi = '{}{}'.format(element, int2roman(charge+1))
        if ion is not None and ion != ioni:
            msg = (
                """
                Inconsistent (element, charge) ({}, {}) vs ion ({})
                """.format(element, charge, ion)
            )
            raise Exception(msg)
        if ION is not None and ION != IONi:
            msg = (
                """
                Inconsistent (element, charge) ({}, {}) vs ION ({})
                """.format(element, charge, ION)
            )
            raise Exception(msg)
        ion = ioni
        ION = IONi

    return ION, ion, element, charge


def _check_elementioncharge_dict(dref_static):
    """ Specific to SpectralLines """

    # Assess if relevant
    lk = [kk for kk in ['ion', 'ION'] if kk in dref_static.keys()]
    if len(lk) == 0:
        return
    kion = lk[0]
    kION = 'ION' if kion == 'ion' else 'ion'
    if kion == 'ION':
        dref_static['ion'] = {}

    lerr = []
    for k0, v0 in dref_static[kion].items():
        try:
            if kion == 'ION':
                ION, ion, element, charge = _check_elementioncharge(
                    ION=k0,
                    ion=v0.get('ion'),
                    element=v0.get('element'),
                    charge=v0.get('charge'),
                )
            else:
                ION, ion, element, charge = _check_elementioncharge(
                    ION=v0.get('ION'),
                    ion=k0,
                    element=v0.get('element'),
                    charge=v0.get('charge'),
                )

            if ION is None:
                continue
            if kion == 'ION':
                dref_static['ion'][ion] = {
                    'ION': ION,
                    'element': element,
                    'charge': charge,
                }
            else:
                dref_static['ion'][k0]['ION'] = ION
                dref_static['ion'][k0]['element'] = element
                dref_static['ion'][k0]['charge'] = charge

        except Exception as err:
            lerr.append((k0, str(err)))

    if kion == 'ION':
        del dref_static['ION']

    if len(lerr) > 0:
        lerr = ['\t- {}: {}'.format(pp[0], pp[1]) for pp in lerr]
        msg = (
            """
            The following entries have non-conform ion / ION / element / charge
            {}
            """.format('\n'.join(lerr))
        )
        raise Exception(msg)


def _harmonize_params(
    dd=None,
    dd_name=None,
    dd_name2=None,
    dref_static=None,
    lkeys=None,
    reserved_keys=None,
    ddefparams=None,
):

    # Check inputs
    if dd_name2 is None:
        dd_name2 = dd_name
    if reserved_keys is None:
        reserved_keys = _DRESERVED_KEYS[dd_name]
    if ddefparams is None:
        ddefparams = _DDEF_PARAMS[dd_name]

    # ------------------
    # list of param keys

    # Get list of known param keys
    lparams = set(itt.chain.from_iterable([
        [k1 for k1 in v0.keys() if k1 not in reserved_keys]
        for k0, v0 in dd.items()
    ]))

    # Add arbitrary params
    if lkeys is not None:
        if isinstance(lkeys, str):
            lkeys = [lkeys]
        c0 = (
            isinstance(lkeys, list)
            and all([isinstance(ss, str) for ss in lkeys])
        )
        if not c0:
            msg = "lkeys must be a list of str!"
            raise Exception(msg)
        lparams = set(lparams).intersection(lkeys)

    # ------------------
    # dparam
    for k0, v0 in ddefparams.items():
        for k1, v1 in dd.items():
            if k0 not in v1.keys():
                dd[k1][k0] = v0[1]
            else:
                # Check type if already included
                if not isinstance(dd[k1][k0], v0[0]):
                    msg = (
                        """
                        Wrong type for parameter:
                            - type({}[{}][{}]) = {}
                        - Expected: {}
                        """.format(
                            dd_name2, k1, k0, type(dd[k1][k0]), v0[0],
                        )
                    )
                    raise Exception(msg)

    for k0 in lparams:
        for k1, v1 in dd.items():
            dd[k1][k0] = dd[k1].get(k0)

    # ------------------
    # Check against dref_static0
    lkpout = [
        (k0, (k1, v0[k1]))
        for k0, v0 in dd.items()
        if k1 in dref_static.keys()
        and any([v0[k1] not in dref_static[k1].keys() for k1 in lparams])
    ]
    if len(lkpout) > 0:
        lpu = sorted(set([pp[1][0] for pp in lkpout]))
        msg0 = '\n'.join([
            '\t- {}[{}]: {}'.format(pp[0], pp[1], pp[2]) for pp in lkpout
        ])
        msg1 = '\n'.join([
            '\t- dref_static[{}]: {}'.format(pp, dref_static[pp].keys())
            for pp in lpu
        ])
        msg = (
            """
            The following parameter have non-identified values in ref_static:
            {}

            Available values:
            {}
            """.format(msg0, msg1)
        )
        raise Exception(msg)

    return dd


def _update_dref_static0(dref_static0=None, ddata0=None, dobj0=None):
    """ Count nb. of matching ref_static in ddata and dobj """

    for k0, v0 in dref_static0.items():

        # ddata
        dd = {
            k2: np.sum([ddata0[k3].get(k0) == k2 for k3 in ddata0.keys()])
            for k2 in v0.keys()
            if any([ddata0[k3].get(k0) == k2 for k3 in ddata0.keys()])
        }
        if len(dd) > 0:
            ss = 'nb. data'
            for k2, v2 in v0.items():
                dref_static0[k0][k2][ss] = int(dd.get(k2, 0))

        # dobj
        for k1, v1 in dobj0.items():
            dd = {
                k2: np.sum([v1[k3].get(k0) == k2 for k3 in v1.keys()])
                for k2 in v0.keys()
                if any([v1[k3].get(k0) == k2 for k3 in v1.keys()])
            }
            if len(dd) > 0:
                ss = 'nb. {}'.format(k1)
                for k2, v2 in v0.items():
                    dref_static0[k0][k2][ss] = int(dd.get(k2, 0))


# #############################################################################
# #############################################################################
#                           Consistency
# #############################################################################


def _consistency(
    dobj=None, dobj0=None,
    ddata=None, ddata0=None,
    dref=None, dref0=None,
    dref_static=None, dref_static0=None,
    dgroup=None, dgroup0=None,
    allowed_groups=None,
    reserved_keys=None,
    ddefparams_data=None,
    ddefparams_obj=None,
    data_none=None,
    max_ndim=None,
):

    # --------------
    # dgroup
    dgroup = _check_dgroup(
        dgroup=dgroup, dgroup0=dgroup0, allowed_groups=allowed_groups,
    )
    dgroup0.update(dgroup)

    # --------------
    # dref
    dref, dgroup_add, ddata_add = _check_dref(
        dref=dref, dref0=dref0, dgroup0=dgroup0, ddata0=ddata0,
        allowed_groups=allowed_groups,
    )
    if dgroup_add is not None:
        dgroup0.update(dgroup_add)
    if ddata_add is not None:
        if ddata is None:
            ddata = ddata_add
        else:
            ddata.update(ddata_add)
    dref0.update(dref)

    # --------------
    # dref_static
    dref_static = _check_dref_static(
        dref_static=dref_static, dref_static0=dref_static0,
    )
    for k0, v0 in dref_static.items():
        if k0 not in dref_static0.keys():
            dref_static0[k0] = v0
        else:
            dref_static0[k0].update(v0)

    # --------------
    # ddata
    ddata, dref_add, dgroup_add = _check_ddata(
        ddata=ddata, ddata0=ddata0,
        dref0=dref0, dgroup0=dgroup0,
        reserved_keys=reserved_keys, allowed_groups=allowed_groups,
        data_none=data_none, max_ndim=max_ndim,
    )
    if dgroup_add is not None:
        dgroup0.update(dgroup_add)
    if dref_add is not None:
        dref0.update(dref_add)
    ddata0.update(ddata)

    # -----------------
    # dobj
    dobj = _check_dobj(
        dobj=dobj, dobj0=dobj0,
    )
    for k0, v0 in dobj.items():
        if k0 not in dobj0.keys():
            dobj0[k0] = v0
        else:
            dobj0[k0].update(v0)

    # --------------
    # params harmonization - ddata
    ddata0 = _harmonize_params(
        dd=ddata0,
        dd_name='ddata',
        dref_static=dref_static0,
        ddefparams=ddefparams_data, reserved_keys=reserved_keys,
    )

    # --------------
    # params harmonization - dobj
    for k0, v0 in dobj0.items():
        dobj0[k0] = _harmonize_params(
            dd=v0,
            dd_name='dobj',
            dd_name2='dobj[{}]'.format(k0),
            dref_static=dref_static0,
            ddefparams=ddefparams_obj.get(k0),
            reserved_keys=reserved_keys,
        )

    # --------------
    # Complement

    # ddata0
    for k0, v0 in ddata0.items():
        if v0.get('data') is None:
            continue
        ddata0[k0]['group'] = tuple([dref0[rr]['group'] for rr in v0['ref']])

    # dref0
    for k0, v0 in dref0.items():
        dref0[k0]['ldata'] = sorted(set(
            k1 for k1 in ddata0.keys()
            if ddata0[k1].get('data') is not None and k0 in ddata0[k1]['ref']
        ))

    # dgroup0
    for k0, v0 in dgroup0.items():
        dgroup0[k0]['lref'] = sorted(set(
            k1 for k1, v1 in dref0.items() if v1['group'] == k0
        ))
        dgroup0[k0]['ldata'] = sorted(set(
            k1 for k1 in ddata0.keys()
            if ddata0[k1].get('data') is not None and k0 in ddata0[k1]['group']
        ))

    # dref_static0
    _update_dref_static0(dref_static0=dref_static0, ddata0=ddata0, dobj0=dobj0)

    # --------------
    # Check conventions
    for k0, v0 in ddata0.items():
        if v0.get('data') is None:
            continue
        if 'time' in v0['group'] and v0['group'].index('time') != 0:
            msg = (
                "ref 'time' must be placed at dimension 0!\n"
                + "\t- ddata['{}']['ref'] = {}\n".format(k0, v0['ref'])
                + "\t- ddata['{}']['group'] = {}".format(k0, v0['group'])
            )
            raise Exception(msg)

    return dgroup0, dref0, dref_static0, ddata0, dobj0


"""
    # --------------
    # params
    lparam = self._ddata['lparam']
    for kk in self._ddata['lkey']:
        for pp in self._ddata['dict'][kk].keys():
            if pp not in self._reserved_all and pp not in lparam:
                lparam.append(pp)
"""


# #############################################################################
# #############################################################################
#               Switch ref
# #############################################################################


def switch_ref(
    new_ref=None,
    ddata=None,
    dref=None,
    dgroup=None,
    dobj0=None,
    dref_static0=None,
    allowed_groups=None,
    reserved_keys=None,
    ddefparams_data=None,
    data_none=None,
    max_ndim=None,
):
    """Use the provided key as ref (if valid) """

    # Check input
    c0 = (
        new_ref in ddata.keys()
        and ddata[new_ref].get('monot') == (True,)
    )
    if not c0:
        strgroup = [
            '{}: {}'.format(
                k0,
                [
                    k1 for k1 in v0['ldata']
                    if ddata[k1].get('monot') == (True,)
                ]
            )
            for k0, v0 in dgroup.items()
        ]
        msg = (
            "\nArg new_ref must be a key to a valid ref (monotonous)!\n"
            + "\t- Provided: {}\n\n".format(new_ref)
            + "Available valid ref candidates:\n"
            + "\t- {}".format('\n\t- '.join(strgroup))
        )
        raise Exception(msg)

    # Substitute in dref
    old_ref = ddata[new_ref]['ref'][0]
    dref[new_ref] = dict(dref[old_ref])
    del dref[old_ref]

    # substitute in ddata['ref']
    for k0, v0 in ddata.items():
        if v0.get('ref') is not None and old_ref in v0['ref']:
            new = tuple([rr for rr in v0['ref']])
            ddata[k0]['ref'] = tuple([
                new_ref if rr == old_ref else rr
                for rr in v0['ref']
            ])

    return _consistency(
        ddata=ddata, ddata0={},
        dref=dref, dref0={},
        dgroup=dgroup, dgroup0={},
        dobj=None, dobj0=dobj0,
        dref_static=None, dref_static0=dref_static0,
        allowed_groups=None,
        reserved_keys=None,
        ddefparams_data=ddefparams_data,
        ddefparams_obj=None,
        data_none=None,
        max_ndim=None,
    )


# #############################################################################
# #############################################################################
#               Get / set / add / remove param
# #############################################################################


def _get_param(
    dd=None, dd_name=None,
    param=None, key=None, ind=None,
    returnas=None,
):
    """ Return the array of the chosen parameter (or list of parameters)

    Can be returned as:
        - dict: {param0: {key0: values0, key1: value1...}, ...}
        - np[.ndarray: {param0: np.r_[values0, value1...], ...}

    """

    # Trivial case
    lp = [kk for kk in list(dd.values())[0].keys() if kk != 'data']
    if param is None:
        param = lp

    # Get key (which data to return param for)
    key = _ind_tofrom_key(dd=dd, key=key, ind=ind, returnas=str)

    # ---------------
    # Check inputs

    # param
    lc = [
        isinstance(param, str) and param in lp and param != 'data',
        isinstance(param, list)
        and all([isinstance(pp, str) and pp in lp for pp in param])
    ]
    if not any(lc):
        msg = (
            "Arg param must a valid param key of a list of such "
            + "(except 'data')\n\n"
            + "Valid params:\n\t- {}\n\n".format('\n\t- '.join(lp))
            + "Provided:\n\t- {}\n".format(param)
        )
        raise Exception(msg)

    if lc[0]:
        param = [param]

    # returnas
    if returnas is None:
        returnas = np.ndarray

    c0 = returnas in [np.ndarray, dict]
    if not c0:
        msg = (
            """
            Arg returnas must be in [np.ndarray, dict]
            Provided: {}
            """.format(returnas)
        )
        raise Exception(msg)

    # -------------
    # Get output

    if returnas == dict:
        out = {k0: {k1: dd[k1][k0] for k1 in key} for k0 in param}
    else:
        out = {k0: np.array([dd[k1][k0] for k1 in key]) for k0 in param}

    return out


def _set_param(
    dd=None, dd_name=None,
    param=None, value=None,
    ind=None, key=None,
):
    """ Set the value of a parameter

    values can be:
        - None
        - a unique value (int, float, bool, str, tuple) => common to all keys
        - an iterable of values (array, list) => one for each key
        - a dict of values (per key)

    A subset of keys can be chosen (ind, key, fed to self.select()) to set
    only the values of some key

    """

    # Check param
    lp = [kk for kk in list(dd.values())[0].keys()]
    if dd_name == 'ddata':
        lp.remove('data')
    if param is None:
        return

    c0 = isinstance(param, str) and param in lp
    if not c0:
        msg = (
            """
            Provided param in not valid
            Valid param:
            {}

            Provided:
            {}
            """.format('\t- ' + '\n\t- '.join(lp), param)
        )
        raise Exception(msg)

    # Check ind / key
    key = _ind_tofrom_key(dd=dd, ind=ind, key=key, returnas='key')

    # Check value
    ltypes = [str, int, np.integer, float, np.floating, tuple]
    lc = [
        isinstance(value, tuple(ltypes)),
        isinstance(value, list) and all([type(tt) in ltypes for tt in value])
        and len(value) == len(key),
        isinstance(value, np.ndarray) and value.shape[0] == len(key),
        isinstance(value, dict)
        and all([
            kk in dd.keys() and type(vv) in ltypes
            for kk, vv in value.items()
        ])
    ]
    if not (value is None or any(lc)):
        msg = (
            """
            Accepted types for value include:
                - None
                - {}: common to all
                - list, np.ndarray: key by key
                - dict of {key: scalar / str}

            The length of value must match the selected keys ({})
            """.format(ltypes, len(key))
        )
        raise Exception(msg)

    # Update data
    if value is None or lc[0]:
        for kk in key:
            dd[kk][param] = value
    elif lc[1] or lc[2]:
        for ii, kk in enumerate(key):
            dd[kk][param] = value[ii]
    else:
        for kk, vv in value.items():
            dd[kk][param] = vv


def _add_param(
    dd=None, dd_name=None,
    param=None, value=None,
):
    """ Add a parameter, optionnally also set its value """
    lp = [kk for kk in list(dd.values())[0].keys()]
    if dd_name == 'ddata':
        lp.remove('data')

    c0 = isinstance(param, str) and param not in lp
    if not c0:
        msg = (
            """
            param must be a str not matching any existing param

            Available param:
            {}

            Provided:
            {}
            """.format(lp, param)
        )
        raise Exception(msg)

    # Initialize and set
    for kk in dd.keys():
        dd[kk][param] = None
    _set_param(dd=dd, param=param, value=value)


def _remove_param(dd=None, dd_name=None, param=None):
    """ Remove a parameter, none by default, all if param = 'all' """

    # Check inputs
    lp = [kk for kk in list(dd.values())[0].keys() if kk != 'data']
    if param is None:
        return
    if param == 'all':
        param = lp

    c0 = isinstance(param, str) and param in lp
    if not c0:
        msg = "Param {} is not a parameter of {}!".format(param, dd_name)
        raise Exception(msg)

    # Remove
    for k0 in dd.keys():
        del dd[k0][param]


# #############################################################################
# #############################################################################
#               Selection
# #############################################################################


def _ind_tofrom_key(
    dd=None, dd_name=None, dgroup=None,
    ind=None, key=None, group=None, returnas=int,
):

    # --------------------
    # Check / format input
    lc = [ind is not None, key is not None]
    if not np.sum(lc) <= 1:
        msg = ("Args ind and key cannot be prescribed simultaneously!")
        raise Exception(msg)

    if dd_name == 'ddata' and group is not None:
        if not (isinstance(group, str) and group in group.keys()):
            msg = (
                """
                Provided group must be valid key of dgroup:
                {}

                Provided:
                {}
                """.format(sorted(dgroup.keys()), group)
            )
            raise Exception(msg)

    lret = [int, bool, str, 'key']
    if returnas not in lret:
        msg = (
            """
            Possible values for returnas are:
            {}

            Provided:
            {}
            """.format(lret, returnas)
        )
        raise Exception(msg)

    # -----------------
    # Compute

    # Intialize output
    out = np.zeros((len(dd),), dtype=bool)

    if not any(lc) and dd_name == 'ddata' and group is not None:
        key = dgroup[group]['ldata']
        lc[1] = True

    # Get output
    lk = list(dd.keys())
    if lc[0]:

        # Check ind
        if not isinstance(ind, np.ndarray):
            ind = np.atleast_1d(ind).ravel()
        c0 = (
            ind.ndim == 1
            and (
                (ind.dtype == np.bool and ind.size == len(dd))
                or (ind.dtype == np.int and ind.size <= len(dd))
            )
        )
        if not c0:
            msg = "Arg ind must be an iterable of bool or int indices!"
            raise Exception(msg)

        # return
        out[ind] = True
        if returnas in [int, str, 'key']:
            out = out.nonzero()[0]
            if returnas in [str, 'key']:
                out = np.array(
                    [kk for ii, kk in enumerate(lk) if ii in out],
                    dtype=str
                )

    elif lc[1]:

        # Check key
        if isinstance(key, str):
            key = [key]
        c0 = (
            isinstance(key, list)
            and all([isinstance(kk, str) and kk in lk for kk in key])
        )
        if not c0:
            msg = (
                """
                key must be valid key to {} (or list of such)
                Provided: {}
                """.format(dd_name, key)
            )
            raise Exception(msg)

        if returnas in ['key', str]:
            out = key
        else:
            for kk in key:
                out[lk.index(kk)] = True
            if returnas == int:
                out = out.nonzero()[0]
    else:
        if returnas == bool:
            out[:] = True
        elif returnas == int:
            out = np.arange(0, len(lk))
        else:
            out = lk
    return out


def _select(dd=None, dd_name=None, log=None, returnas=None, **kwdargs):
    """ Return the indices / keys of data matching criteria

    The selection is done comparing the value of all provided parameters
    The result is a boolean indices array, optionally with the keys list
    It can include:
        - log = 'all': only the data matching all criteria
        - log = 'any': the data matching any criterion

    If log = 'raw', a dict of indices arrays is returned, showing the
    details for each criterion

    """

    # Format and check input
    if log is None:
        log = 'all'
    if returnas is None:
        returnas = bool if log == 'raw' else int
    if log not in ['all', 'any', 'raw']:
        msg = (
            "Arg log must be:\n"
            + "\t- 'all': all criteria should match\n"
            + "\t- 'any': any criterion should match\n"
            + "\t- 'raw': return the full 2d array of boolean indices\n\n"
            + "Provided:\n\t{}".format(log)
        )
        raise Exception(msg)
    if returnas not in [int, bool, str, 'key']:
        msg = (
            "Arg returnas must be:\n"
            + "\t- bool: array of boolean indices\n"
            + "\t- int: array of int indices\n"
            + "\t- str / 'key': array of keys\n\n"
            + "Provided:\n\t{}".format(returnas)
        )
        raise Exception(msg)

    kwdargs = {k0: v0 for k0, v0 in kwdargs.items() if v0 is not None}

    # Get list of relevant criteria
    lp = [kk for kk in list(dd.values())[0].keys()]
    if dd_name == 'ddata':
        lp.remove('data')

    lcritout = [ss for ss in kwdargs.keys() if ss not in lp]
    if len(lcritout) > 0:
        msg = (
            """
            The following criteria correspond to no parameters:
                - {}
              => only use known parameters (self.dparam_{}.keys()):
                - {}
            """.format(lcritout, dd_name, '\n\t- '.join(lp))
        )
        raise Exception(msg)

    # Prepare array of bool indices and populate
    ltypes = [float, np.float_]
    lquant = [
        kk for kk in kwdargs.keys()
        if any([type(dd[k0][kk]) in ltypes for k0 in dd.keys()])
    ]

    ind = np.zeros((len(kwdargs), len(dd)), dtype=bool)
    for ii, kk in enumerate(kwdargs.keys()):
        try:
            par = _get_param(
                dd=dd, dd_name=dd_name,
                param=kk,
                returnas=np.ndarray,
            )[kk]
            if kk in lquant:
                if isinstance(kwdargs[kk], list) and len(kwdargs[kk]) == 2:
                    ind[ii, :] = (
                        (kwdargs[kk][0] <= par) & (par <= kwdargs[kk][1])
                    )
                elif isinstance(kwdargs[kk], tuple) and len(kwdargs[kk]) == 2:
                    ind[ii, :] = (
                        (kwdargs[kk][0] > par) | (par > kwdargs[kk][1])
                    )
                else:
                    ind[ii, :] = par == kwdargs[kk]
            else:
                ind[ii, :] = par == kwdargs[kk]
        except Exception as err:
            try:
                ind[ii, :] = [
                    dd[k0][kk] == kwdargs[kk] for k0 in dd.keys()
                ]
            except Exception as err:
                msg = (
                    "Could not determine whether:\n"
                    + "\t- {}['{}'] == {}".format(
                        dd_name, kk, kwdargs[kk],
                    )
                )
                raise Exception(msg)

    # Format output ind
    if log == 'raw':
        if returnas in [str, 'key']:
            ind = {
                kk: [k0 for jj, k0 in enumerate(dd.keys()) if ind[ii, jj]]
                for ii, kk in enumerate(kwdargs.keys())
            }
        if returnas == int:
            ind = {
                kk: ind[ii, :].nonzero()[0]
                for ii, kk in enumerate(kwdargs.keys())
            }
        else:
            ind = {kk: ind[ii, :] for ii, kk in enumerate(kwdargs.keys())}
    else:
        if log == 'all':
            ind = np.all(ind, axis=0)
        else:
            ind = np.any(ind, axis=0)
        if returnas == int:
            ind = ind.nonzero()[0]
        elif returnas in [str, 'key']:
            ind = np.array(
                [k0 for jj, k0 in enumerate(dd.keys()) if ind[jj]],
                dtype=str,
            )
    return ind


def _get_keyingroup_ddata(
    dd=None, dd_name='data',
    key=None, group=None, monot=None,
    msgstr=None, raise_=False,
):
    """ Return the unique data key matching key in desired group in ddata

    Here, key can be interpreted as name / source / units / quant...
    All are tested using select() and a unique match is returned
    If not unique match an error message is either returned or raised

    """

    # ------------------------
    # Trivial case: key is actually a ddata key
    if key in dd.keys():
        lg = dd[key]['group']
        if group is None or group in lg:
            return key, None
        else:
            msg = ("Required data key does not have matching group:\n"
                   + "\t- {}['{}']['group'] = {}\n".format(dd_name, key, lg)
                   + "\t- Expected group:  {}".format(group))
            if raise_:
                raise Exception(msg)

    # ------------------------
    # Non-trivial: check for a unique match on other params

    dind = _select(
        dd=dd, dd_name=dd_name,
        dim=key, quant=key, name=key, units=key, source=key,
        group=group, monot=monot,
        log='raw', returnas=bool,
    )
    ind = np.array([ind for kk, ind in dind.items() if kk != 'group'])
    if group is not None:
        ind &= dind['group'][None, :]

    # Any perfect match ?
    nind = np.sum(ind, axis=1)
    sol = (nind == 1).nonzero()[0]
    key_out, msg = None, None
    if sol.size > 0:
        if np.unique(sol).size == 1:
            indkey = ind[sol[0], :].nonzero()[0]
            key_out = list(dd.keys())[indkey]
        else:
            lstr = "[dim, quant, name, units, source]"
            msg = "Several possible matches in {} for {}".format(lstr, key)
    else:
        lstr = "[dim, quant, name, units, source]"
        msg = "No match in {} for {} in group {}".format(lstr, key, group)

    # Complement error msg and optionally raise
    if msg is not None:
        lk = ['dim', 'quant', 'name', 'units', 'source']
        dk = {
            kk: (
                dind[kk].sum(),
                sorted(set([vv[kk] for vv in dd.values()]))
            ) for kk in lk
        }
        msg += (
            "\n\nRequested {} could not be identified!\n".format(msgstr)
            + "Please provide a valid (unique) key/name/dim/quant/units:\n\n"
            + '\n'.join([
                '\t- {} ({} matches): {}'.format(kk, dk[kk][0], dk[kk][1])
                for kk in lk
            ])
            + "\nProvided:\n\t'{}'".format(key)
        )
        if raise_:
            raise Exception(msg)
    return key_out, msg


def _get_possible_ref12d(
    dd=None,
    key=None, ref1d=None, ref2d=None,
    group1d='radius',
    group2d='mesh2d',
):

    # Get relevant lists
    kq, msg = _get_keyingroup_ddata(
        dd=dd,
        key=key, group=group2d, msgstr='quant', raise_=False,
    )

    if kq is not None:
        # The desired quantity is already 2d
        k1d, k2d = None, None

    else:
        # Check if the desired quantity is 1d
        kq, msg = _get_keyingroup_ddata(
            dd=dd,
            key=key, group=group1d,
            msgstr='quant', raise_=True,
        )

        # Get dict of possible {ref1d: lref2d}
        ref = [rr for rr in dd[kq]['ref'] if dd[rr]['group'] == (group1d,)][0]
        lref1d = [
            k0 for k0, v0 in dd.items()
            if ref in v0['ref'] and v0['monot'][v0['ref'].index(ref)] is True
        ]

        # Get matching ref2d with same quant and good group
        lquant = list(set([dd[kk]['quant'] for kk in lref1d]))
        dref2d = {
            k0: [
                kk for kk in _select(
                    dd=dd, quant=dd[k0]['quant'],
                    log='all', returnas=str,
                )
                if group2d in dd[kk]['group']
                and not isinstance(dd[kk]['data'], dict)
            ]
            for k0 in lref1d
        }
        dref2d = {k0: v0 for k0, v0 in dref2d.items() if len(v0) > 0}

        if len(dref2d) == 0:
            msg = (
                "No match for (ref1d, ref2d) for ddata['{}']".format(kq)
            )
            raise Exception(msg)

        # check ref1d
        if ref1d is None:
            if ref2d is not None:
                lk = [k0 for k0, v0 in dref2d.items() if ref2d in v0]
                if len(lk) == 0:
                    msg = (
                        "\nNon-valid interpolation intermediate\n"
                        + "\t- provided:\n"
                        + "\t\t- ref1d = {}, ref2d = {}\n".format(ref1d, ref2d)
                        + "\t- valid:\n{}".format(
                            '\n'.join([
                                '\t\t- ref1d = {}  =>  ref2d in {}'.format(
                                    k0, v0
                                )
                                for k0, v0 in dref2d.items()
                            ])
                        )
                    )
                    raise Exception(msg)
                if kq in lk:
                    ref1d = kq
                else:
                    ref1d = lk[0]
            else:
                if kq in dref2d.keys():
                    ref1d = kq
                else:
                    ref1d = list(dref2d.keys())[0]
        else:
            ref1d, msg = _get_keyingroup_ddata(
                dd=dd,
                key=ref1d, group=group1d,
                msgstr='ref1d', raise_=False,
            )
        if ref1d not in dref2d.keys():
            msg = (
                "\nNon-valid interpolation intermediate\n"
                + "\t- provided:\n"
                + "\t\t- ref1d = {}, ref2d = {}\n".format(ref1d, ref2d)
                + "\t- valid:\n{}".format(
                    '\n'.join([
                        '\t\t- ref1d = {}  =>  ref2d in {}'.format(
                            k0, v0
                        )
                        for k0, v0 in dref2d.items()
                    ])
                )
            )
            raise Exception(msg)

        # check ref2d
        if ref2d is None:
            ref2d = dref2d[ref1d][0]
        else:
            ref2d, msg = _get_keyingroup_ddata(
                dd=dd,
                key=ref2d, group=group2d,
                msgstr='ref2d', raise_=False,
            )
        if ref2d not in dref2d[ref1d]:
            msg = (
                "\nNon-valid interpolation intermediate\n"
                + "\t- provided:\n"
                + "\t\t- ref1d = {}, ref2d = {}\n".format(ref1d, ref2d)
                + "\t- valid:\n{}".format(
                    '\n'.join([
                        '\t\t- ref1d = {}  =>  ref2d in {}'.format(
                            k0, v0
                        )
                        for k0, v0 in dref2d.items()
                    ])
                )
            )
            raise Exception(msg)

    return kq, ref1d, ref2d





# Built-in
import os
import itertools as itt

# Common
import numpy as np
from scipy.interpolate import BSpline
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
import matplotlib.patches as mpatches
import matplotlib.colors as mcolors
from matplotlib.colors import ListedColormap
import matplotlib.gridspec as gridspec
from matplotlib.axes._axes import Axes
from mpl_toolkits.mplot3d import Axes3D

# tofu
from tofu.version import __version__

__all__ = ['plot_fit1d']


_GITHUB = 'https://github.com/ToFuProject/tofu/issues'
_WINTIT = 'tofu-%s        report issues / requests at %s' % (
                __version__, _GITHUB
                )

_QUIVERCOLOR = plt.cm.viridis(np.linspace(0, 1, 3))
_QUIVERCOLOR = np.array([[1., 0., 0., 1.],
                         [0., 1., 0., 1.],
                         [0., 0., 1., 1.]])
_QUIVERCOLOR = ListedColormap(_QUIVERCOLOR)


# #################################################################
# #################################################################
#                   data plot
# #################################################################
# #################################################################


def CrystalBragg_plot_data_vs_lambphi(
    xi, xj, bragg, lamb, phi, data,
    lambfit=None, phifit=None,
    spect1d=None, vertsum1d=None,
    lambax=None, phiax=None,
    phiminmax=None, dlines=None,
    lambmin=None, lambmax=None,
    xjcut=None, lambcut=None,
    phicut=None, spectcut=None,
    cmap=None, vmin=None, vmax=None,
    fs=None, dmargin=None,
    tit=None, wintit=None,
    angunits='deg',
):

    # Check inputs
    # ------------

    if fs is None:
        fs = (14, 8)
    if tit is None:
        tit = False
    if wintit is None:
        wintit = _WINTIT
    if cmap is None:
        cmap = plt.cm.viridis
    if dmargin is None:
        dmargin = {'left': 0.05, 'right': 0.99,
                   'bottom': 0.07, 'top': 0.92,
                   'wspace': 0.2, 'hspace': 0.3}
    assert angunits in ['deg', 'rad']
    if angunits == 'deg':
        bragg = bragg*180./np.pi
        phi = phi*180./np.pi
        phifit = phifit*180./np.pi
        if phiax is not None:
            phiax = 180*phiax/np.pi
        phiminmax = phiminmax*180./np.pi
        if phicut is not None:
            phicut = phicut*180./np.pi
    nspect = spect1d.shape[0]

    if dlines is not None:
        lines = [k0 for k0, v0 in dlines.items()
                 if (v0['lambda0'] >= lambfit[0]
                     and v0['lambda0'] <= lambfit[-1])]
        lions = sorted(set([dlines[k0]['ION'] for k0 in lines]))
        nions = len(lions)
        dions = {k0: [k1 for k1 in lines if dlines[k1]['ION'] == k0]
                 for k0 in lions}
        dions = {k0: {'lamb': np.array([dlines[k1]['lambda0']
                                        for k1 in dions[k0]]),
                      'symbol': [dlines[k1]['symbol'] for k1 in dions[k0]]}
                 for k0 in lions}
        lcol = ['k', 'r', 'b', 'g', 'm', 'c']
        ncol = len(lcol)
        llamb = np.concatenate([dions[lions[ii]]['lamb']
                                for ii in range(nions)])
        indsort = np.argsort(llamb)
        lsymb = list(itt.chain.from_iterable([dions[lions[ii]]['symbol']
                                              for ii in range(nions)]))
        lsymb = [lsymb[ii] for ii in indsort]
        lcollamb = list(itt.chain.from_iterable(
            [[lcol[ii]]*dions[lions[ii]]['lamb'].size
             for ii in range(nions)]))
        lcollamb = [lcollamb[ii] for ii in indsort]

    lcolspect = ['k', 'r', 'b', 'g', 'y', 'm', 'c']
    ncolspect = len(lcolspect)

    # pre-compute
    # ------------

    # extent
    extent = (xi.min(), xi.max(), xj.min(), xj.max())
    if lambmin is None:
        lambmin = lambfit.min()
    if lambmax is None:
        lambmax = lambfit.max()

    # Plot
    # ------------

    fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(4, 4, **dmargin)
    ax0 = fig.add_subplot(gs[:3, 0], aspect='equal')   # , adjustable='datalim'
    ax1 = fig.add_subplot(
        gs[:3, 1],
        aspect='equal',
        sharex=ax0, sharey=ax0,
    )    # , adjustable='datalim'
    axs1 = fig.add_subplot(gs[3, 1], sharex=ax0)
    ax2 = fig.add_subplot(gs[:3, 2])
    axs2 = fig.add_subplot(gs[3, 2], sharex=ax2, sharey=axs1)
    ax3 = fig.add_subplot(gs[:3, 3], sharey=ax2)

    ax0.set_title('Coordinates transform')
    ax1.set_title('Camera image')
    axs1.set_title('Vertical average')
    ax2.set_title('Camera image transformed')

    ax0.set_ylabel(r'$x_j$ (m)')
    ax0.set_xlabel(r'$x_i$ (m)')
    axs1.set_ylabel(r'data')
    axs1.set_xlabel(r'$x_i$ (m)')
    ax2.set_ylabel(r'incidence angle ($deg$)')
    axs2.set_xlabel(r'$\lambda$ ($m$)')

    ax0.contour(xi, xj, bragg, 10, cmap=cmap)
    ax0.contour(xi, xj, phi, 10, cmap=cmap, ls='--')
    ax1.imshow(data, extent=extent, aspect='equal',
               origin='lower', vmin=vmin, vmax=vmax)
    axs1.plot(xi, np.nanmean(data, axis=0), c='k', ls='-')
    ax2.scatter(lamb.ravel(), phi.ravel(), c=data.ravel(), s=2,
                marker='s', edgecolors='None',
                cmap=cmap, vmin=vmin, vmax=vmax)

    if xjcut is None:
        ax3.plot(vertsum1d, phifit, c='k', ls='-')
    else:
        dphicut = 0.1*(phifit.max() - phifit.min())
        for ii in range(xjcut.size):
            ax1.axhline(xjcut[ii],
                        ls='-', lw=1., c='r')
            ax2.plot(lambcut[ii, :], phicut[ii, :],
                     ls='-', lw=1., c='r')
            ax3.plot(lambcut[ii, :],
                     (np.nanmean(phicut[ii, :])
                      + spectcut[ii, :]*dphicut/np.nanmax(spectcut[ii, :])),
                     ls='-', lw=1., c='r')

    if phiax is not None:
        ax2.plot(lambax, phiax, c='r', ls='-', lw=1.)

    # Plot known spectral lines
    if dlines is not None:
        for ii, k0 in enumerate(lions):
            for jj in range(dions[k0]['lamb'].size):
                x = dions[k0]['lamb'][jj]
                col = lcol[ii % ncol]
                axs2.axvline(x,
                             c=col, ls='--')
                axs2.annotate(dions[k0]['symbol'][jj],
                              xy=(x, 1.01), xytext=None,
                              xycoords=('data', 'axes fraction'),
                              color=col, arrowprops=None,
                              horizontalalignment='center',
                              verticalalignment='bottom')
                if xjcut is not None:
                    ax3.axvline(x,
                                c=col, ls='--')
        hand = [mlines.Line2D([], [], color=lcol[ii % ncol], ls='--')
                for ii in range(nions)]
        axs2.legend(hand, lions,
                    bbox_to_anchor=(1., 1.02), loc='upper left')

    # Plot 1d spectra and associated phi windows
    for ii in range(nspect):
        axs2.plot(
            lambfit, spect1d[ii],
            c=lcolspect[ii % ncolspect],
            ls='-',
            )
        ax2.axhline(
            phiminmax[ii, 0],
            c=lcolspect[ii % ncolspect],
            ls='-', lw=1.,
            )
        ax2.axhline(
            phiminmax[ii, 1],
            c=lcolspect[ii % ncolspect],
            ls='-', lw=1.,
            )
        ax3.axhline(
            phiminmax[ii, 0],
            c=lcolspect[ii % ncolspect],
            ls='-', lw=1.,
            )
        ax3.axhline(
            phiminmax[ii, 1],
            c=lcolspect[ii % ncolspect],
            ls='-', lw=1.,
            )

    ax2.set_xlim(lambmin, lambmax)
    ax2.set_ylim(phifit.min(), phifit.max())
    plt.setp(ax1.get_xticklabels(), visible=False)
    plt.setp(ax1.get_yticklabels(), visible=False)
    plt.setp(ax2.get_xticklabels(), visible=False)
    plt.setp(ax3.get_yticklabels(), visible=False)

    if tit is not False:
        fig.suptitle(tit, size=14, weight='bold')
    if wintit is not False:
        fig.canvas.set_window_title(wintit)
    return [ax0, ax1]


# #################################################################
# #################################################################
#                   fit1d plot
# #################################################################
# #################################################################


def plot_fit1d(
    dfit1d=None, dextract=None, annotate=None, showonly=None,
    indspect=None, xlim=None, fs=None, dmargin=None,
    tit=None, wintit=None,
):

    # Check inputs
    # ------------
    if annotate is None:
        annotate = True
    if annotate is True:
        annotate = dfit1d['dinput']['keys']
    if isinstance(annotate, str):
        annotate = [annotate]
    if xlim is None:
        xlim = False
    if fs is None:
        fs = (15, 8)
    if wintit is None:
        wintit = _WINTIT
    if dmargin is None:
        dmargin = {'left': 0.05, 'right': 0.85,
                   'bottom': 0.07, 'top': 0.85,
                   'wspace': 0.2, 'hspace': 0.3}

    # Extract (better redeability)
    dprepare = dfit1d['dinput']['dprepare']
    dinput = dfit1d['dinput']

    # Index of spectra to plot
    if not np.any(dinput['valid']['indt']):
        msg = "The provided fit1d result has no valid time step!"
        raise Exception(msg)

    if indspect is None:
        indspect = dinput['valid']['indt'].nonzero()[0][0]
    indspect = np.atleast_1d(indspect).ravel()
    if indspect.dtype == 'bool':
        indspect = np.nonzero(indspect)[0]

    # pre-compute
    # ------------
    if dinput['same_spectrum'] is True:
        nlines = int(dinput['nlines'] / dinput['same_spectrum_nspect'])
        dinput['lines'] = dinput['lines'][:nlines]
        dinput['ion'] = dinput['ion'][:nlines]
        dinput['symb'] = dinput['symb'][:nlines]
        nwidth = int(dinput['width']['keys'].size
                     / dinput['same_spectrum_nspect'])
        dinput['width']['keys'] = dinput['width']['keys'][:nwidth]
        dinput['width']['ind'] = dinput['width']['ind'][:, :nlines]
        dinput['width']['ind'] = dinput['width']['ind'][:nwidth, :]
        dinput['shift']['ind'] = dinput['shift']['ind'][:, :nlines]

    nlines = dinput['lines'].size
    ions_u = sorted(set(dinput['ion'].tolist()))
    nions = len(ions_u)

    indwidth = np.argmax(dinput['width']['ind'], axis=0)
    indshift = np.argmax(dinput['shift']['ind'], axis=0)

    x = dinput['lines'][None, :] + dextract['shift']['values']

    lcol = ['k', 'r', 'b', 'g', 'm', 'c']
    ncol = len(lcol)
    if dextract['Ti'] is not False:
        lfcol = ['y', 'g', 'c', 'm']
    else:
        lfcol = [None]
    nfcol = len(lfcol)
    if dextract['vi'] is not False:
        lhatch = [None, '/', '\\', '|', '-', '+', 'x', '//']
    else:
        lhatch = [None]
    nhatch = len(lhatch)
    nspect = indspect.size

    # Plot
    # ------------

    for ii in range(nspect):
        ispect = indspect[ii]
        if tit is None:
            titi = ("spect {}\n".format(ispect+1)
                    + "({}/{})".format(ispect, dprepare['data'].shape[0]))
        else:
            titi = tit

        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(1, 1, **dmargin)
        ax = fig.add_subplot(gs[0, 0])
        ax.set_ylabel(r'data (a.u.)')
        ax.set_xlabel(r'$\lambda$ (m)')

        ax.plot(dprepare['lamb'][dprepare['indok'][ispect, :]],
                dprepare['data'][ispect, dprepare['indok'][ispect, :]],
                marker='.', c='k', ls='None', ms=8)
        if showonly is not True:
            if dextract['sol_detail'] is not False:
                ax.plot(
                    dprepare['lamb'], dextract['sol_detail'][ispect, :, 0],
                    ls='-', c='k',
                )
            ax.set_prop_cycle(None)
            if dextract['sol_detail'] is not False:
                if dextract['Ti'] is not False or dextract['vi'] is not False:
                    for jj in range(nlines):
                        col = lfcol[indwidth[jj] % nfcol]
                        hatch = lhatch[indshift[jj] % nhatch]
                        ax.fill_between(
                            dprepare['lamb'],
                            dextract['sol_detail'][ispect, :, 1+jj],
                            alpha=0.3, color=col, hatch=hatch,
                        )
                else:
                    ax.plot(
                        dprepare['lamb'],
                        dextract['sol_detail'][ispect, :, 1:].T,
                    )
            if dextract['sol_tot'] is not False:
                ax.plot(
                    dprepare['lamb'],
                    dextract['sol_tot'][ispect, :],
                    c='k', lw=2.,
                )

        # Annotate lines
        if annotate is not False:
            for jj, k0 in enumerate(ions_u):
                col = lcol[jj % ncol]
                ind = (dinput['ion'] == k0).nonzero()[0]
                for nn in ind:
                    if dinput['keys'][nn] not in annotate:
                        continue
                    lab = dinput['symb'][nn]
                    ax.axvline(x[ispect, nn], c=col, ls='--')
                    if dextract['coefs'] is not False:
                        val = dextract['coefs']['values'][ispect, nn]
                        lab += '\n{:4.2e}'.format(val)
                    if dextract['shift'] is not False:
                        val = dextract['shift']['values'][ispect, nn]*1.e10
                        lab += '\n({:+4.2e} A)'.format(val)
                    ax.annotate(
                        lab,
                        xy=(x[ispect, nn], 1.01), xytext=None,
                        xycoords=('data', 'axes fraction'),
                        color=col, arrowprops=None,
                        horizontalalignment='center',
                        verticalalignment='bottom',
                    )

            # Ion legend
            hand = [
                mlines.Line2D(
                    [], [],
                    color=lcol[jj % ncol], ls='--', label=ions_u[jj]
                )
                for jj in range(nions)
            ]
            legi = ax.legend(
                handles=hand,
                title='ions',
                bbox_to_anchor=(1.01, 1.),
                loc='upper left',
            )
            ax.add_artist(legi)

        # Ti legend
        if dextract['Ti'] is not False:
            hand = [mpatches.Patch(color=lfcol[jj % nfcol])
                    for jj in range(dinput['width']['ind'].shape[0])]
            lleg = [str(dinput['width']['keys'][jj])
                    + '{:4.2f}'.format(
                            dextract['Ti']['values'][ispect, jj]*1.e-3
                            )
                    for jj in range(dinput['width']['ind'].shape[0])]
            legT = ax.legend(
                handles=hand, labels=lleg,
                title='Ti (keV)',
                bbox_to_anchor=(1.01, 1.),    # 0.8
                loc='upper left',
            )
            ax.add_artist(legT)

        # vi legend
        if dextract['vi'] is not False:
            hand = [mpatches.Patch(facecolor='w', edgecolor='k',
                                   hatch=lhatch[jj % nhatch])
                    for jj in range(dinput['shift']['ind'].shape[0])]
            lleg = [str(dinput['shift']['keys'][jj])
                    + '{:4.2f}'.format(
                            dextract['vi']['values'][ispect, jj]*1.e-3
                            )
                    for jj in range(dinput['shift']['ind'].shape[0])]
            legv = ax.legend(
                handles=hand, labels=lleg,
                title='vi (km/s)',
                bbox_to_anchor=(1.01, 0.75),    # 0.5
                loc='upper left',
            )
            ax.add_artist(legv)

        # Ratios legend
        if dextract['ratio'] is not False:
            nratio = dextract['ratio']['values'].shape[1]
            hand = [mlines.Line2D([], [], c='k', ls='None')]*nratio
            lleg = ['{} =  {:4.2e}'.format(
                dextract['ratio']['lab'][jj],
                dextract['ratio']['values'][ispect, jj])
                    for jj in range(nratio)]
            legr = ax.legend(
                handles=hand,
                labels=lleg,
                title='line ratio',
                bbox_to_anchor=(1.01, 0.30),    # 0.21
                loc='lower left',
            )
            ax.add_artist(legr)

        # bck legend
        if True:
            hand = [mlines.Line2D([], [], c='k', ls='None')]*2
            lleg = [
                'amp = {:4.2e}'.format(dextract['bck_amp']['values'][ispect]),
                'rate = {:4.2e}'.format(dextract['bck_rate']['values'][ispect])
            ]
            legr = ax.legend(
                handles=hand,
                labels=lleg,
                title='background',
                bbox_to_anchor=(1.01, 0.10),    # 0.05
                loc='lower left',
            )
            ax.add_artist(legr)

        # double legend
        if dinput['double'] is not False:
            hand = [mlines.Line2D([], [], c='k', ls='None')]*2
            lleg = [
                'ratio = {:4.2f}'.format(dextract['dratio'][ispect]),
                'shift ' + r'$\approx$'
                + ' {:4.2e}'.format(dextract['dshift'][ispect])
            ]
            legr = ax.legend(
                handles=hand,
                labels=lleg,
                title='double',
                bbox_to_anchor=(1.01, -0.1),
                loc='lower left',
            )

        ax.set_xlim(dinput['dprepare']['domain']['lamb']['minmax'])

        if titi is not False:
            fig.suptitle(titi, size=14, weight='bold')
        if wintit is not False:
            fig.canvas.set_window_title(wintit)
    if xlim is not False:
        ax.set_xlim(xlim)
    return ax


def CrystalBragg_plot_data_fit2d(xi, xj, data, lamb, phi, indok=None,
                                 dfit2d=None,
                                 dax=None, indspect=None,
                                 spect1d=None, fit1d=None,
                                 lambfit=None, phiminmax=None,
                                 cmap=None, vmin=None, vmax=None,
                                 plotmode=None, fs=None, dmargin=None,
                                 angunits='deg', tit=None, wintit=None):

    # Check inputs
    # ------------

    if dax is None:
        if fs is None:
            fs = (18, 9)
        if cmap is None:
            cmap = plt.cm.viridis
        if dmargin is None:
            dmargin = {'left': 0.05, 'right': 0.99,
                       'bottom': 0.06, 'top': 0.95,
                       'wspace': None, 'hspace': 0.8}
        if wintit is None:
            wintit = _WINTIT
        if tit is None:
            tit = '2D fitting of X-Ray Crystal Bragg spectrometer'
    if dfit2d['dinput']['symmetry'] is True:
        symaxis = dfit2d['dinput']['symmetry_axis'][indspect]
    if angunits is None:
        angunits = 'deg'
    assert angunits in ['deg', 'rad']

    phiflat = dfit2d['phi']
    pts_phi = dfit2d['pts_phi']
    ylim = np.r_[dfit2d['dinput']['phiminmax']]
    if angunits == 'deg':
        phiflat = phiflat*180./np.pi
        phi = phi*180./np.pi
        pts_phi = pts_phi*180./np.pi
        ylim = ylim*180./np.pi
        phiminmax = phiminmax*180./np.pi
        if dfit2d['dinput']['symmetry'] is True:
            symaxis = symaxis*180./np.pi
    if plotmode == 'transform':
        xlab = r'$\lambda$ (m)'
        ylab = r'$\phi$ ({})'.format(angunits)
    else:
        xlab = r'$x_i$' + r' (m)'
        ylab = r'$x_j$' + r' ({})'.format(angunits)
        spect1d = None

    # pre-compute
    # ------------

    if vmin is None:
        vmin = 0.
    if vmax is None:
        vmax = max(np.nanmax(dfit2d['data'][indspect, :]),
                   np.nanmax(dfit2d['sol_tot'][indspect, :]))
    err = dfit2d['sol_tot'][indspect, :] - dfit2d['data'][indspect, :]
    errm = vmax/10.

    # Prepare figure if dax not provided
    # ------------
    if dax is None:
        fig = plt.figure(figsize=fs)
        naxh = (3*3
                + 2*(dfit2d['dinput']['Ti'] + dfit2d['dinput']['vi']
                     + (dfit2d['ratio'] not in [None, False])))
        naxv = 1
        gs = gridspec.GridSpec((3+naxv)*3, naxh, **dmargin)
        if plotmode == 'transform':
            ax0 = fig.add_subplot(gs[:9, :3])
            ax0.set_xlim(dfit2d['dinput']['lambminmax'])
            ax0.set_ylim(ylim)
        else:
            ax0 = fig.add_subplot(gs[:9, :3], aspect='equal')
        ax0c = fig.add_subplot(gs[10, :3])
        ax1 = fig.add_subplot(
            gs[:9, 3:6],
            sharex=ax0, sharey=ax0,
        )
        ax2 = fig.add_subplot(
            gs[:9, 6:9],
            sharex=ax0, sharey=ax0,
        )
        ax1c = fig.add_subplot(gs[10, 3:6])

        ax0.set_title('Residu')
        ax0.set_xlabel(xlab)
        ax0.set_ylabel(ylab)
        ax1.set_title('Original data')
        ax1.set_xlabel(xlab)
        ax2.set_title('2d fit')
        dax = {'err': ax0, 'err_colorbar': ax0c,
               'data': ax1, 'data_colorbar': ax1c,
               'fit': ax2}
        if tit is not False:
            fig.suptitle(tit)
        if wintit is not False:
            fig.canvas.set_window_title(wintit)

        if naxv == 1:
            dax['fit1d'] = fig.add_subplot(gs[9:11, 6:9], sharex=ax2)
            dax['fit1d'].set_title('1d sliced spectrum vs fit')
            dax['fit1d'].set_ylabel(r'data')
            dax['err1d'] = fig.add_subplot(gs[11, 6:9], sharex=ax2)
            dax['err1d'].set_ylabel(r'err')
            dax['err1d'].set_xlabel(xlab)

        nn = 1
        if dfit2d['dinput']['Ti'] is True:
            dax['Ti'] = fig.add_subplot(gs[:9, 9+2*(nn-1):9+2*nn], sharey=ax1)
            dax['Ti'].set_title(r'Width')
            dax['Ti'].set_xlabel(r'$\hat{T_i}$' + r' (keV)')
            nn += 1
        if dfit2d['dinput']['vi'] is True:
            dax['vi'] = fig.add_subplot(gs[:9, 9+2*(nn-1):9+2*nn], sharey=ax1)
            dax['vi'].set_title(r'Shift')
            dax['vi'].set_xlabel(r'$\hat{v_i}$' + r' (km/s)')
            dax['vi'].axvline(0, ls='-', lw=1., c='k')
            nn += 1
        if dfit2d['ratio'] not in [None, False]:
            dax['ratio'] = fig.add_subplot(gs[:9, 9+2*(nn-1):], sharey=ax1)
            dax['ratio'].set_title(r'Intensity Ratio')
            dax['ratio'].set_xlabel(r'ratio (a.u)')

    # Plot main images
    # ------------
    if plotmode == 'transform':
        if dax.get('err') is not None:
            errax = dax['err'].scatter(dfit2d['lamb'],
                                       phiflat,
                                       c=err,
                                       s=6, marker='s', edgecolors='None',
                                       vmin=-errm, vmax=errm,
                                       cmap=plt.cm.seismic)
        if dax.get('data') is not None:
            dataax = dax['data'].scatter(dfit2d['lamb'],
                                         phiflat,
                                         c=dfit2d['data'][indspect, :],
                                         s=6, marker='s', edgecolors='None',
                                         vmin=vmin, vmax=vmax, cmap=cmap)
        if dax.get('fit') is not None:
            dax['fit'].scatter(dfit2d['lamb'],
                               phiflat,
                               c=dfit2d['sol_tot'][indspect, :],
                               s=6, marker='s', edgecolors='None',
                               vmin=vmin, vmax=vmax, cmap=cmap)
            if dfit2d['dinput']['symmetry'] is True:
                dax['fit'].axhline(symaxis,
                                   c='k', ls='--', lw=2.)
        if dax.get('fit1d') is not None:
            # dax['fit1d'].plot()
            pass
    else:
        extent = (xi.min(), xi.max(), xj.min(), xj.max())
        if dax.get('err') is not None:
            dax['err'].imshow(err,
                              extent=extent, cmap=plt.cm.seismic,
                              origin='lower', vmin=errm, vmax=errm)
        if dax.get('data') is not None:
            dax['data'].imshow(data,
                               extent=extent, cmap=cmap,
                               origin='lower', vmin=vmin, vmax=vmax)
        if dax.get('fit') is not None:
            dax['fit'].imshow(fit,
                              extent=extent, cmap=cmap,
                              origin='lower', vmin=vmin, vmax=vmax)

    # Plot profiles
    # ------------
    if dax.get('Ti') is not None and dfit2d['dinput']['Ti'] is True:
        for ii in range(dfit2d['kTiev'].shape[1]):
            dax['Ti'].plot(dfit2d['kTiev'][indspect, ii, :]*1.e-3,
                           pts_phi,
                           ls='-', marker='.', ms=4,
                           label=dfit2d['dinput']['width']['keys'][ii])
        dax['Ti'].set_xlim(left=0.)
        dax['Ti'].legend(frameon=True,
                         loc='upper left', bbox_to_anchor=(0., -0.1))

    if dax.get('vi') is not None and dfit2d['dinput']['vi'] is True:
        for ii in range(dfit2d['vims'].shape[1]):
            dax['vi'].plot(dfit2d['vims'][indspect, ii, :]*1.e-3,
                           pts_phi,
                           ls='-', marker='.', ms=4,
                           label=dfit2d['dinput']['shift']['keys'][ii])
        dax['vi'].legend(frameon=True,
                         loc='upper left', bbox_to_anchor=(0., -0.1))

    if dax.get('ratio') not in [None, False]:
        for ii in range(dfit2d['ratio']['value'].shape[1]):
            dax['ratio'].plot(dfit2d['ratio']['value'][indspect, ii, :],
                              pts_phi,
                              ls='-', marker='.', ms=4,
                              label=dfit2d['ratio']['str'][ii])
        dax['ratio'].legend(frameon=True,
                            loc='upper left', bbox_to_anchor=(0., -0.1))

    # Add 1d prof
    # ------------
    if spect1d is not None:
        lcol = ['r', 'k', 'm']
        for ii in range(spect1d.shape[0]):
            l, = dax['fit1d'].plot(lambfit, fit1d[ii, :],
                                   ls='-', lw=1., c=lcol[ii % len(lcol)])
            col = l.get_color()
            dax['fit1d'].plot(lambfit, spect1d[ii, :],
                              ls='None', marker='.', c=col, ms=4)
            dax['err1d'].plot(lambfit, fit1d[ii, :] - spect1d[ii, :],
                              ls='-', lw=1., c=col)
            dax['fit'].axhline(phiminmax[ii][0], c=col, ls='-', lw=2.)
            dax['fit'].axhline(phiminmax[ii][1], c=col, ls='-', lw=2.)

    # double legend
    if dfit2d['dinput']['double'] is not False:
        hand = [mlines.Line2D([], [], c='k', ls='None')]*2
        c0 = (dfit2d['dinput']['double'] is True
              or dfit2d['dinput']['double'].get('dratio') is None)
        if c0:
            dratio = dfit2d['dratio'][indspect]
            dratiostr = ''
        else:
            dratio = dfit2d['dinput']['double']['dratio']
            dratiostr = '  (fixed)'
        c0 = (dfit2d['dinput']['double'] is True
              or dfit2d['dinput']['double'].get('dshift') is None)
        if c0:
            dshift = dfit2d['dshift'][indspect]
            dshiftstr = ''
        else:
            dshift = dfit2d['dinput']['double']['dshift']
            dshiftstr = '  (fixed)'
        lleg = ['dratio = {:4.2f}{}'.format(dratio, dratiostr),
                ('dshift = {:4.2e} * '.format(dshift)
                 + r'$\lambda${}'.format(dshiftstr))]
        legr = dax['err1d'].legend(handles=hand, labels=lleg, title='double',
                                   bbox_to_anchor=(1.01, 0.),
                                   loc='center left')

    # Polishing
    # ------------
    for kax in dax.keys():
        if kax != 'err':
            plt.setp(dax[kax].get_yticklabels(), visible=False)
    if dax.get('fit') is not None:
        plt.setp(dax['fit'].get_xticklabels(), visible=False)
    if dax.get('fit1d') is not None:
        plt.setp(dax['fit1d'].get_xticklabels(), visible=False)
    if dax.get('err_colorbar') is not None and dax['err'] is not None:
        plt.colorbar(errax, ax=dax['err'], cax=dax['err_colorbar'],
                     orientation='horizontal', extend='both')
    if (
        dax.get('data_colorbar') is not None
        and (dax['data'] is not None or dax['fit'] is not None)
    ):
        plt.colorbar(dataax, ax=dax['data'], cax=dax['data_colorbar'],
                     orientation='horizontal')
    return dax


# #################################################################
# #################################################################
#                   noise plot
# #################################################################
# #################################################################

def plot_noise_analysis(dnoise=None, margin=None, fraction=None,
                        ms=None, dcolor=None,
                        dax=None, fs=None, dmargin=None,
                        wintit=None, tit=None, sublab=None,
                        save=None, name=None, path=None, fmt=None):

    # Check inputs
    # ------------
    if save is None:
        save = False

    if fraction is None:
        fraction = 0.4

    if ms is None:
        ms = 4.
    if dcolor is None:
        dcolor = {'mask': (0.4, 0.4, 0.4),
                  'noeval': (0.7, 0.7, 0.7),
                  'S/N': (1., 0., 0.)}
    dcolor['ok'] = dcolor.get('ok', (0., 0., 0.))

    if dax is None:
        if fs is None:
            fs = (18, 9)
        if dmargin is None:
            dmargin = {'left': 0.05, 'right': 0.99,
                       'bottom': 0.06, 'top': 0.95,
                       'wspace': 0.5, 'hspace': None}
        if wintit is None:
            wintit = _WINTIT
        if tit is None:
            tit = '2D fitting of X-Ray Crystal Bragg spectrometer'
        if sublab is None:
            sublab = True

    # Prepare data
    # ------------
    nbs = dnoise['nbsplines']
    nspect = dnoise['data'].shape[0]
    nlamb = dnoise['chi2n'].shape[1]

    indsort = dnoise['indsort']
    indnan = dnoise['indnan'][:-1]
    phiplot = np.tile(np.insert(dnoise['phi'][indsort[0, :], indsort[1, :]],
                                indnan,
                                np.nan),
                      (nspect, 1)).ravel()
    dataplot = np.insert(dnoise['data'][:, indsort[0, :], indsort[1, :]],
                         indnan,
                         np.nan, axis=1).ravel()
    fitplot = np.insert(dnoise['fit'][:, indsort[0, :], indsort[1, :]],
                        indnan,
                        np.nan, axis=1).ravel()
    errplot = fitplot - dataplot

    # get all indices
    indin = dnoise['indin']
    indout_mask = dnoise['indout_mask']
    indout_noeval = dnoise['indout_noeval']

    # fit sqrt on sigmai
    c0 = (dnoise['var_fraction'] is not False
          or (margin is not None and margin != dnoise['var_margin'])
          or (fraction is not None
              and fraction != dnoise['var_fraction']))
    if c0 is True:
        from . import _fit12d
        (mean, var, xdata, const,
         indout_var, _, margin, _) = _fit12d.get_noise_analysis_var_mask(
             fit=dnoise['fit'], data=dnoise['data'], mask=dnoise['mask'],
             margin=margin, fraction=False)
    else:
        mean = dnoise['var_mean']
        var = dnoise['var']
        const = dnoise['var_const']
        xdata = dnoise['var_xdata']
        indout_var = dnoise['indout_var']
        assert dnoise['var_fraction'] is False

    # Safety check
    indout_var_agg = (np.sum(indout_var, axis=0)/float(indout_var.shape[0])
                      > fraction)
    indout_tot = np.array([indout_mask,
                           indout_noeval,
                           indout_var_agg])
    c0 = np.all(np.sum(indout_tot.astype(int), axis=0) <= 1)
    if not c0:
        msg = "Overlapping indout!"
        raise Exception(msg)

    # get plotting indout
    indout_maskplot = np.tile(np.insert(
        indout_mask[indsort[0, :], indsort[1, :]], indnan, False),
        (nspect, 1)).ravel()
    indout_noevalplot = np.tile(np.insert(
        indout_noeval[indsort[0, :], indsort[1, :]], indnan, False),
        (nspect, 1)).ravel()
    indout_varplot = np.insert(
        indout_var[:, indsort[0, :], indsort[1, :]],
        indnan, False, axis=1).ravel()
    dindout = {
        'mask': {'ind': indout_mask, 'plot': indout_maskplot},
        'noeval': {'ind': indout_noeval},
        # ,'plot': {indout_noevalplot},
        'S/N': {'ind': indout_var_agg, 'plot': indout_varplot},
    }
    indinplot = ~(indout_maskplot | indout_noevalplot | indout_varplot)

    # cam and cmap
    cam = np.zeros(indout_mask.shape, dtype=float)
    for ii, k0 in enumerate(['mask', 'noeval', 'S/N']):
        cam[dindout[k0]['ind']] = ii+1
    cam = cam/4

    cmap = plt.cm.get_cmap('viridis', 4)
    newcolors = cmap(np.linspace(0, 1, 4))
    for ii, k0 in enumerate(['ok', 'mask', 'noeval', 'S/N']):
        newcolors[ii, :] = mcolors.to_rgba(dcolor[k0])
    cmap = ListedColormap(newcolors)

    alpha = 0.2

    # Prepare figure if dax not provided
    # ------------
    if dax is None:
        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(2, 8, **dmargin)
        ax0 = fig.add_subplot(gs[0, 2:4])
        ax1 = fig.add_subplot(gs[0, 4:6], sharey=ax0)
        ax2 = fig.add_subplot(gs[1, 2:4])
        ax3 = fig.add_subplot(gs[1, 4:6])
        ax4 = fig.add_subplot(gs[:, 6:], aspect='equal', adjustable='datalim')
        ax5 = fig.add_subplot(gs[0, 1], sharey=ax0)
        ax6 = fig.add_subplot(gs[1, :2])

        ax0.set_title('Profiles')
        ax0.set_xlabel(r'data (a.u.)')
        ax1.set_title('Err profiles')
        ax1.set_xlabel(r'data (a.u.)')
        ax2.set_title('Error')
        ax2.set_xlabel('fit (a.u.)')
        ax2.set_ylabel('err (a.u.)')
        ax3.set_title('Standard deviation')
        ax3.set_xlabel('data (a.u.)')
        ax3.set_ylabel(r'$\sigma$' + ' (a.u.)')
        ax4.set_title('Camera')
        ax4.set_xlabel(r'$x_i$ (m)')
        ax4.set_ylabel(r'$x_j$ (m)')
        ax5.set_title('bsplines')
        ax5.set_xlabel('')
        ax5.set_ylabel(r'$\phi$ (rad)')
        ax6.set_title('')
        ax6.set_xlabel('mean(fit)')
        ax6.set_ylabel(r'$\chi^2_n$ (a.u.)')

        dax = {'bsplines': ax5,
               'prof': ax0,
               'proferr': ax1,
               'err': ax2,
               'errbin': ax3,
               'cam': ax4,
               'chi2': ax6,
               }

    # Plot main images
    # ------------

    # bsplines
    if dax.get('bsplines') is not None:
        dax['bsplines'].plot(dnoise['bs_val'],
                             np.repeat(dnoise['bs_phi'][:, None],
                                       dnoise['nbsplines'], axis=1),
                             ls='-', lw=1.)

    # Profile
    if dax.get('prof') is not None:
        dax['prof'].plot(dataplot[indinplot], phiplot[indinplot],
                         marker='.', ls='None',
                         mfc=(0., 0., 0., alpha), mec='None', ms=ms)
        dax['prof'].plot(fitplot, phiplot,
                         marker='None', ls='-', lw=1., c='g')
        for k0 in ['mask', 'S/N']:
            dax['prof'].plot(dataplot[dindout[k0]['plot']],
                             phiplot[dindout[k0]['plot']],
                             marker='.', ls='None',
                             mfc=np.r_[dcolor[k0], alpha], mec='None', ms=ms)

    # Profile err
    if dax.get('proferr') is not None:
        dax['proferr'].plot(errplot[indinplot], phiplot[indinplot],
                            marker='.', ls='None',
                            mfc=(0., 0., 0., alpha), mec='None', ms=ms)
        for k0 in ['mask', 'S/N']:
            dax['proferr'].plot(errplot[dindout[k0]['plot']],
                                phiplot[dindout[k0]['plot']],
                                marker='.', ls='None',
                                mfc=np.r_[dcolor[k0], alpha],
                                mec='None', ms=ms)

    # Error distribution
    if dax.get('err') is not None:
        dax['err'].axhline(0., c='k', ls='--', lw=1.)
        lab = r'{}'.format(margin) + r'$\sigma$'
        dax['err'].fill_between(xdata,
                                -margin*const*np.sqrt(xdata),
                                margin*const*np.sqrt(xdata),
                                color=(0.7, 0.7, 0.7, 1.), label=lab)
        dax['err'].plot(fitplot[indinplot], errplot[indinplot],
                        marker='.', ls='None',
                        mfc=(0., 0., 0., alpha), mec='None', ms=ms)
        for k0 in ['mask', 'S/N']:
            dax['err'].plot(fitplot[dindout[k0]['plot']],
                            errplot[dindout[k0]['plot']],
                            marker='.', ls='None',
                            mfc=np.r_[dcolor[k0], alpha], mec='None', ms=ms)
        dax['err'].legend(loc='lower right', frameon=True)

    # Error binning and standard variation estimate
    if dax.get('errbin') is not None:
        dax['errbin'].plot(xdata, np.sqrt(var),
                           marker='.', c='k', ls='None')
        lab = r'$\sigma=$' + r'{:5.3e}'.format(const) + r'$\sqrt{data}$'
        dax['errbin'].plot(xdata, const*np.sqrt(xdata),
                           marker='None', ls='-', c='k', label=lab)
        dax['errbin'].plot(xdata, np.sqrt(xdata),
                           marker='None', ls='--', c='k')
        dax['errbin'].plot(xdata, mean,
                           marker='.', ls='None', c='k')
        dax['errbin'].axhline(0., ls='--', c='k', lw=1., marker='None')
        dax['errbin'].legend(loc='lower right', frameon=True)

    # Camera with identified pixels
    if dax.get('cam') is not None:
        dax['cam'].imshow(cam.T,
                          origin='lower', interpolation='nearest',
                          aspect='equal', cmap=cmap)
        lab = r'err > {:02.0f}% {}$\sigma$'.format(fraction*100, margin)
        hand = [mlines.Line2D([], [], c=dcolor['S/N'], marker='s', ls='None')]
        dax['cam'].legend(hand, [lab], loc='best')

    # chi2
    if dax.get('chi2') is not None:
        dax['chi2'].plot(dnoise['chi2_meandata'].ravel(),
                         (dnoise['chi2n']).ravel(),
                         c='k', marker='.', ms=ms, ls='None')

    # Polish
    # ------------
    if sublab is True:
        if dax.get('bsplines') is not None:
            dax['bsplines'].annotate('(a)',
                                     xy=(0., 1.02),
                                     xycoords='axes fraction',
                                     size=10, fontweight='bold',
                                     horizontalalignment='center',
                                     verticalalignment='bottom')
        if dax.get('prof') is not None:
            dax['prof'].annotate('(b)',
                                 xy=(0., 1.02),
                                 xycoords='axes fraction',
                                 size=10, fontweight='bold',
                                 horizontalalignment='center',
                                 verticalalignment='bottom')
        if dax.get('proferr') is not None:
            dax['proferr'].annotate('(c)',
                                    xy=(0., 1.02),
                                    xycoords='axes fraction',
                                    size=10, fontweight='bold',
                                    horizontalalignment='center',
                                    verticalalignment='bottom')
        if dax.get('chi2') is not None:
            dax['chi2'].annotate('(d)',
                                 xy=(0., 1.02),
                                 xycoords='axes fraction',
                                 size=10, fontweight='bold',
                                 horizontalalignment='center',
                                 verticalalignment='bottom')
        if dax.get('err') is not None:
            dax['err'].annotate('(e)',
                                xy=(0., 1.02),
                                xycoords='axes fraction',
                                size=10, fontweight='bold',
                                horizontalalignment='center',
                                verticalalignment='bottom')
        if dax.get('errbin') is not None:
            dax['errbin'].annotate('(f)',
                                   xy=(0., 1.02),
                                   xycoords='axes fraction',
                                   size=10, fontweight='bold',
                                   horizontalalignment='center',
                                   verticalalignment='bottom')
        if dax.get('cam') is not None:
            dax['cam'].annotate('(g)',
                                xy=(0., 1.02),
                                xycoords='axes fraction',
                                size=10, fontweight='bold',
                                horizontalalignment='center',
                                verticalalignment='bottom')

    # save
    # ------------
    if save is True:
        if name is None:
            name = 'NoiseAnalysis'
        if fmt is None:
            fmt = 'png'
        if path is None:
            path = './'
        if name[-4:] != '.{}'.format(fmt):
            name = '{}.{}'.format(name, fmt)

        pfe = os.path.join(os.path.abspath(path), name)
        dax['prof'].figure.savefig(pfe, format=fmt)

        msg = "Saved in:\n\t{}".format(pfe)
        print(msg)
    return dax


def plot_noise_analysis_scannbs(
    dnoise=None, ms=None,
    dax=None, fs=None, dmargin=None,
    wintit=None, tit=None, sublab=None,
    save=None, name=None, path=None, fmt=None,
):

    # Check inputs
    # ------------
    if save is None:
        save = False

    if ms is None:
        ms = 2.

    if dax is None:
        if fs is None:
            fs = (18, 9)
        if dmargin is None:
            dmargin = {'left': 0.06, 'right': 0.99,
                       'bottom': 0.06, 'top': 0.96,
                       'wspace': 0.5, 'hspace': 1.}
        if wintit is None:
            wintit = _WINTIT
        if tit is None:
            tit = '2D fitting of X-Ray Crystal Bragg spectrometer'
        if sublab is None:
            sublab = True

    # Prepare data
    # ------------
    nlbs = dnoise['lnbsplines'].size
    nbs = dnoise['nbsplines']
    nspect = dnoise['chi2n'].shape[1]
    nlambu = dnoise['chi2n'].shape[2]

    # Plotting arrays
    lnbsplines_plot = np.repeat(dnoise['lnbsplines'], nspect*nlambu)
    chi2n_plot = dnoise['chi2n'].ravel()
    alpha = np.tile((np.abs(dnoise['dataint'])
                     / np.nanmax(np.abs(dnoise['dataint']))),
                    (nlbs, 1, 1)).ravel()
    colors = np.concatenate((np.tile([0., 0., 0.], (nspect*nlambu*nlbs, 1)),
                             alpha[:, None]),
                            axis=1)
    chi2nmean = (np.nansum(dnoise['chi2n']*dnoise['dataint'][None, :, :],
                           axis=-1).sum(axis=-1)
                 / np.nansum(dnoise['dataint']))
    var_mean = np.sqrt(np.nansum(dnoise['var_mean']**2, axis=1))
    var_std = np.sqrt(np.nansum(
        (np.sqrt(dnoise['var'])
         - dnoise['var_const'][:, None]*np.sqrt(dnoise['var_xdata']))**2,
        axis=1))

    # Prepare figure if dax not provided
    # ------------
    if dax is None:
        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(11, max(1, dnoise['nbsplines'].size*4+2),
                               **dmargin)
        ax0 = fig.add_subplot(gs[:4, :])
        ax0.set_xlabel(r'nb. of bsplines')
        ax0.set_ylabel(r'Residue (a.u.)')
        ax0.set_title('nlamb = {}'.format(dnoise['lambedges'].size-1))

        dax = {'conv': ax0,
               'cases_fit': [],
               'cases_err': [],
               'cases_var': []}
        for ii in range(dnoise['nbsplines'].size):
            if ii == 0:
                dax['cases_fit'].append(fig.add_subplot(
                    gs[5:9, ii*5:ii*5+2]))
                dax['cases_err'].append(fig.add_subplot(
                    gs[5:9, ii*5+2:ii*5+4],
                    sharey=dax['cases_fit'][0]))
                dax['cases_var'].append(fig.add_subplot(
                    gs[9:, ii*5:ii*5+4]))
            else:
                dax['cases_fit'].append(fig.add_subplot(
                    gs[5:9, ii*5:ii*5+2],
                    sharey=dax['cases_fit'][ii-1]))
                dax['cases_err'].append(fig.add_subplot(
                    gs[5:9, ii*5+2:ii*5+4],
                    sharex=dax['cases_err'][0],
                    sharey=dax['cases_fit'][0]))
                dax['cases_var'].append(fig.add_subplot(
                    gs[9:, ii*5:ii*5+4],
                    sharey=dax['cases_var'][ii-1]))
            # dax['cases_err'][ii].
            dax['cases_fit'][ii].set_ylabel(r'$\phi$ (rad)')
            dax['cases_fit'][ii].set_xlabel(r'data (a.u.)')
            dax['cases_err'][ii].set_xlabel(r'error (a.u.)')
            dax['cases_var'][ii].set_xlabel(r'data (a.u.)')
            dax['cases_var'][ii].set_ylabel(r'$\sigma$ (a.u.)')

            dax['cases_fit'][ii].xaxis.set_label_coords(0.5, -0.07)
            dax['cases_err'][ii].xaxis.set_label_coords(0.5, -0.07)
            plt.setp(dax['cases_err'][ii].get_yticklabels(), visible=False)

    # Plot main images
    # ------------

    # Fit chi2n vs nbs
    if dax.get('conv') is not None:
        minus_one = chi2nmean[-1]
        chi2nn = ((dnoise['var_const'][0] - dnoise['var_const'][-1])
                  * (chi2nmean - minus_one) / (chi2nmean[0] - minus_one)
                  + dnoise['var_const'][-1])
        dax['conv'].plot(dnoise['lnbsplines'], chi2nn,
                         c='k', ls='-', lw=2., label='fit cost')

        # error dictribution moments vs nbs
        varmn = ((dnoise['var_const'][0] - dnoise['var_const'][-1])
                 * (var_mean - var_mean[-1]) / (var_mean[0] - var_mean[-1])
                 + dnoise['var_const'][-1])
        dax['conv'].plot(dnoise['lnbsplines'], varmn,
                         c='b', ls='-', lw=2., label=r'$\hat{err}_{cost}$')
        varstdn = ((dnoise['var_const'][0] - dnoise['var_const'][-1])
                   * (var_std - var_std[-1]) / (var_std[0] - var_std[-1])
                   + dnoise['var_const'][-1])
        dax['conv'].plot(dnoise['lnbsplines'], varstdn,
                         c='r', ls='-', lw=2., label=r'$\sigma_{cost}$')
        dax['conv'].plot(dnoise['lnbsplines'], dnoise['var_const'],
                         c='g', ls='-', lw=2., label='fitting constant')
        dax['conv'].legend(loc='best')

    # Particular cases
    if dax.get('case_fit') is not None and dnoise.get('nbsplines') is not None:
        inbs = [np.nonzero(dnoise['lnbsplines'] == nbs)[0][0]
                for nbs in dnoise['nbsplines']]
        for ii in range(dnoise['nbsplines'].size):
            indin = dnoise['bs_indin'][ii]
            dax['cases_fit'][ii].plot(dnoise['bs_data'][ii][indin],
                                      dnoise['bs_phidata'][ii][indin],
                                      c=(0.5, 0.5, 0.5),
                                      marker='.', ls='None', ms=ms)
            dax['cases_fit'][ii].plot(dnoise['bs_fit'][ii][indin],
                                      dnoise['bs_phidata'][ii][indin],
                                      c='k', marker='None', ls='-', lw=2.)

            err = (dnoise['bs_fit'][ii] - dnoise['bs_data'][ii])
            dax['cases_err'][ii].scatter(
                err[indin], dnoise['bs_phidata'][ii][indin],
                marker='.', facecolors=(0., 0., 0., 0.5),
                edgecolors='None')
            dax['cases_err'][ii].axvline(0., c='k', ls='--', lw=1.)

            # var
            indok = ~np.isnan(dnoise['var'][inbs[ii], :])
            varfit = (dnoise['var_const'][inbs[ii]]
                      * np.sqrt(dnoise['var_xdata'][indok]))
            lab = (r"{:4.2f}".format(dnoise['var_const'][inbs[ii]])
                   + r"$\sqrt{data}$")
            dax['cases_var'][ii].fill_between(
                dnoise['var_xdata'][indok],
                varfit, np.sqrt(dnoise['var'][inbs[ii], indok]),
                fc=(1., 0., 0., 0.5), ls='-')
            dax['cases_var'][ii].plot(
                dnoise['var_xdata'],
                np.sqrt(dnoise['var'][inbs[ii], :]),
                c='r', marker='.', ls='None', ms=ms)
            dax['cases_var'][ii].plot(
                dnoise['var_xdata'][indok], varfit,
                c='g', marker='None', ls='-', label=lab)

            # mean
            dax['cases_var'][ii].fill_between(
                dnoise['var_xdata'], 0.,
                dnoise['var_mean'][inbs[ii], :],
                fc=(0., 0., 1., 0.5), ls='-')
            dax['cases_var'][ii].axhline(0., c='k', ls='--', lw=1.)
            dax['cases_var'][ii].plot(dnoise['var_xdata'],
                                      dnoise['var_mean'][inbs[ii], :],
                                      c='b', ls='-', lw=1.)
            dax['cases_var'][ii].legend(loc='upper left')

            posx = 1.1
            posy = 1.05
            plt.text(posx, posy, '{} bsplines'.format(dnoise['nbsplines'][ii]),
                     horizontalalignment='center',
                     verticalalignment='center',
                     size=12, fontweight='bold',
                     transform=dax['cases_fit'][ii].transAxes)

            if dax.get('conv') is not None:
                dax['conv'].axvline(dnoise['nbsplines'][ii],
                                    c='k', ls='--', lw=1.)

    # Polish
    # ------------
    if sublab is True:
        if dax.get('conv') is not None:
            dax['conv'].annotate('(a)',
                                 xy=(0., 1.02),
                                 xycoords='axes fraction',
                                 size=10, fontweight='bold',
                                 horizontalalignment='center',
                                 verticalalignment='bottom')
        if dax.get('cases_fit') is not None:
            ls = ['b', 'c', 'd', 'e', 'f']
            for ii in range(len(dax['cases_fit'])):
                dax['cases_fit'][ii].annotate(
                    '({})'.format(ls[ii]),
                    xy=(0., 1.02), xycoords='axes fraction',
                    size=10, fontweight='bold',
                    horizontalalignment='center',
                    verticalalignment='bottom')

    # save
    # ------------
    if save is True:
        if name is None:
            name = 'NoiseAnalysis'
        if fmt is None:
            fmt = 'png'
        if path is None:
            path = './'
        if name[-4:] != '.{}'.format(fmt):
            name = '{}.{}'.format(name, fmt)

        pfe = os.path.join(os.path.abspath(path), name)
        dax['conv'].figure.savefig(pfe, format=fmt)

        msg = "Saved in:\n\t{}".format(pfe)
        print(msg)
    return dax



"""
This module contains tests for tofu.geom in its structured version
"""

# Built-in
import os
import shutil
import itertools as itt
import warnings


# Standard
import numpy as np
import matplotlib.pyplot as plt


# tofu-specific
from tofu import __version__
import tofu as tf
import tofu.data as tfd


_here = os.path.abspath(os.path.dirname(__file__))
VerbHead = 'tofu.mesh.test_01_checks'
_TOFU_USER = os.path.join(os.path.expanduser("~"), '.tofu')
_CUSTOM = os.path.dirname(os.path.dirname(os.path.dirname(_here)))
_CUSTOM = os.path.join(_CUSTOM, 'scripts', 'tofucustom.py')


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module():
    print("Removing user ~/.tofu/ if any")
    if os.path.isdir(_TOFU_USER):
        shutil.rmtree(_TOFU_USER)
    # Recreating clean .tofu
    # out = subprocess.run(_CUSTOM, stdout=PIPE, stderr=PIPE)
    os.system('python '+_CUSTOM)


def teardown_module():
    print("Removing user ~/.tofu/ if any")
    if os.path.isdir(_TOFU_USER):
        shutil.rmtree(_TOFU_USER)


#######################################################
#
#     checking routines
#
#######################################################


class Test01_checks():

    @classmethod
    def setup_class(cls):
        pass

    @classmethod
    def setup(self):
        pass

    def teardown(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test01_mesh2DRect_X_check(self):

        lx = [[1, 2], [1, 2, 3, 4]]
        lres = [None, 10, 0.1, [0.1, 0.2], [0.1, 0.2, 0.3, 0.1]]

        for comb in itt.product(lx, lres):
            if hasattr(lres, '__iter__') and len(lres) != len(lx):
                continue
            x, res, ind = tfd._mesh_checks._mesh2DRect_X_check(
                x=[1, 2, 3, 4],
                res=10,
            )
            if hasattr(lres, '__iter__'):
                assert x_new.size == np.unique(x_new).size == res.size + 1


#######################################################
#
#     object mesh2D
#
#######################################################


class Test02_Mesh2DRect():

    @classmethod
    def setup_class(cls):
        pass

    def setup(self):
        self.dobj = {
            'm0': tfd.Mesh2DRect(),
            'm1': tfd.Mesh2DRect(),
            'm2': tfd.Mesh2DRect(),
            'm3': None,
        }

        # add mesh
        ldomain = [
            [[2, 3], [-1, 1]],
            [[2, 2.3, 2.6, 3], [-1, 0., 1]],
            [[2, 3], [-1, 0, 1]],
        ]
        lres = [
            0.1,
            [[0.2, 0.1, 0.1, 0.2], [0.2, 0.1, 0.2]],
            [0.1, [0.2, 0.1, 0.2]],
        ]

        for ii, (k0, v0) in enumerate(self.dobj.items()):
            if k0 != 'm3':
                self.dobj[k0].add_mesh(
                    domain=ldomain[ii],
                    res=lres[ii],
                    key=k0,
                )
            else:
                self.dobj[k0] = tfd.Mesh2DRect.from_Config(
                    tf.load_config('WEST'),
                    res=0.1,
                    key=k0,
                )

        # add splines
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            self.dobj[k0].add_bsplines(deg=ii)

    def teardown(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test01_get_summary(self):
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            self.dobj[k0].get_summary()

    def test02_select_ind(self):
        lkey = ['m0', 'm1-bs1', 'm2', 'm3-bs3']
        lelements = ['cents', None, 'knots', None]
        lind = [None, ([0, 5], [0, 6]), [0, 10, 100], ([0, 5, 6], [0, 2, 3])]
        lcrop = [True, False, True, False]
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            indt = self.dobj[k0].select_ind(
                key=lkey[ii],
                ind=lind[ii],
                elements=lelements[ii],
                returnas=tuple,
                crop=lcrop[ii],
            )
            indf = self.dobj[k0].select_ind(
                key=lkey[ii],
                ind=indt,
                elements=lelements[ii],
                returnas=np.ndarray,
                crop=lcrop[ii],
            )
            indt2 = self.dobj[k0].select_ind(
                key=lkey[ii],
                ind=indf,
                elements=lelements[ii],
                returnas=tuple,
                crop=lcrop[ii],
            )
            assert all([np.allclose(indt[ii], indt2[ii]) for ii in [0, 1]])

    def test03_select_mesh(self):
        lkey = ['m0', 'm1', 'm2', 'm3']
        lind = [None, ([0, 5], [0, 6]), [0, 10, 100], ([0, 5, 6], [0, 2, 3])]
        lelements = ['cents', 'cents', 'knots', None]
        lreturnas = ['ind', 'data', 'data', 'ind']
        lreturn_neig = [None, True, False, True]
        lcrop = [False, True, False, False]
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            indf = self.dobj[k0].select_mesh_elements(
                key=lkey[ii],
                ind=lind[ii],
                elements=lelements[ii],
                returnas=lreturnas[ii],
                return_neighbours=lreturn_neig[ii],
                crop=lcrop[ii],
            )

    def test04_select_bsplines(self):
        lkey = ['m0-bs0', 'm1-bs1', 'm2-bs2', 'm3-bs3']
        lind = [None, ([0, 5], [0, 6]), [0, 10, 100], ([0, 5, 6], [0, 2, 3])]
        lreturnas = [None, 'data', 'data', 'ind']
        lreturn_cents = [None, True, False, True]
        lreturn_knots = [None, False, True, True]
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            indf = self.dobj[k0].select_bsplines(
                key=lkey[ii],
                ind=lind[ii],
                returnas=lreturnas[ii],
                return_cents=lreturn_cents[ii],
                return_knots=lreturn_knots[ii],
            )

    def test05_sample_mesh(self):
        lres = [None, 0.1, 0.01, [0.1, 0.05]]
        lmode = [None, 'rel', 'abs', 'abs']
        lgrid = [None, True, False, False]
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            out = v0.get_sample_mesh(
                res=lres[ii], grid=lgrid[ii], mode=lmode[ii],
            )

    """
    def test06_sample_bspline(self):
        lres = [None, 0.1, 0.01, [0.1, 0.05]]
        lmode = [None, 'rel', 'abs', 'abs']
        lgrid = [None, True, False, False]
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            out = v0.get_sample_bspline(
                res=lres[ii], grid=lgrid[ii], mode=lmode[ii],
            )
    """

    def test07_plot_mesh(self):
        lik = [None, ([0, 2], [0, 3]), [2, 3], None]
        lic = [None, ([0, 2], [0, 3]), None, [2, 3]]
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            dax = self.dobj[k0].plot_mesh(
                ind_knot=lik[ii],
                ind_cent=lic[ii],
            )
        plt.close('all')

    def test08_plot_bsplines(self):
        lkey = ['m0-bs0', 'm1-bs1', 'm2-bs2', 'm3-bs3']
        lind = [None, ([1, 2], [2, 1]), (1, 1), [1, 2, 10]]
        lknots = [None, True, False, True]
        lcents = [False, False, True, True]
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            dax = self.dobj[k0].plot_bsplines(
                key=lkey[ii],
                ind=lind[ii],
                knots=lknots[ii],
                cents=lcents[ii],
            )
        plt.close('all')

    def test09_plot_profile2d(self):
        lkey = ['m0-bs0', 'm1-bs1', 'm2-bs2', 'm3-bs3']
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            key = str(ii)
            kbs = lkey[ii]
            ref = self.dobj[k0].dobj['bsplines'][kbs]['ref']
            shapebs = self.dobj[k0].dobj['bsplines'][kbs]['shape']

            self.dobj[k0].add_data(
                key=key,
                data=np.random.random(shapebs),
                ref=ref,
            )

            dax = self.dobj[k0].plot_profile2d(
                key=key,
            )
        plt.close('all')

    def test10_add_bsplines_operator(self):
        lkey = ['m0-bs0', 'm1-bs1', 'm2-bs2']
        lop = ['D0', 'D0N2', 'D1N2', 'D2N2']
        lgeom = ['linear', 'toroidal']
        lcrop = [False, True]
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            if ii == 3:
                continue

            for comb in itt.product(lop, lgeom, lcrop):
                deg = self.dobj[k0].dobj['bsplines'][lkey[ii]]['deg']
                if int(comb[0][1]) > deg:
                    continue
                self.dobj[k0].add_bsplines_operator(
                    key=lkey[ii],
                    operator=comb[0],
                    geometry=comb[1],
                    crop=comb[2],
                )

    def test11_compute_plot_geometry_matrix(self):

        # get config and cam
        conf = tf.load_config('WEST-V0')
        cam = tf.geom.utils.create_CamLOS1D(
            pinhole=[3., 1., 0.],
            orientation=[np.pi, 0., 0],
            focal=0.1,
            sensor_nb=50,
            sensor_size=0.15,
            config=conf,
            Diag='SXR',
            Exp='WEST',
            Name='cam1',
        )

        # compute geometry matrices
        for ii, (k0, v0) in enumerate(self.dobj.items()):

            mat = self.dobj[k0].compute_geometry_matrix(
                cam=cam, res=0.01, crop=True,
            )

            dax = mat.plot_geometry_matrix(cam=cam, indchan=12, indbf=100)
            plt.close('all')
